from . import world
from .constants import const, Constants
from .core import Arg, From, FromArg, Get, inject, Inject, Provide, Scope, wire, Wiring
from .factory import Factory, factory
from .implementation import implementation
from .lazy import LazyCall, LazyMethodCall
from .service import ABCService, Service, service
from .utils import is_compiled

try:
    from ._internal.scm_version import version as __version__
except ImportError:  # pragma: no cover
    __version__ = ''

__all__ = ['__version__', 'world', 'Get', 'From', 'FromArg', 'Provide', 'Inject',
           'const', 'Constants', 'constants', 'inject', 'Arg', 'wire', 'Wiring', 'factory',
           'Factory', 'implementation', 'LazyCall', 'LazyMethodCall',
           'service', 'Scope', 'Service', 'ABCService', 'is_compiled']



# pyright: reportUnknownMemberType=false
from __future__ import annotations

from typing import cast, Dict, Tuple, Type

from .slots import SlotsRepr
from .. import API


@API.private
class ImmutableMeta(type):
    def __new__(mcs: Type[ImmutableMeta],
                name: str,
                bases: Tuple[type, ...],
                namespace: Dict[str, object],
                **kwargs: object
                ) -> ImmutableMeta:
        if '__slots__' not in namespace:
            raise TypeError("Attributes must be defined in slots")

        # TODO: Type ignore necessary when type checking with Python 3.6
        #       To be removed ASAP.
        return super().__new__(mcs, name, bases, namespace, **kwargs)


@API.private
class Immutable(SlotsRepr, metaclass=ImmutableMeta):
    """
    Imitates immutable behavior by raising an exception when modifying an
    attribute through the standard means.
    """
    __slots__ = ()

    def __init__(self, *args: object, **kwargs: object) -> None:
        # quick way to initialize an Immutable through args. It won't take into
        # account parent classes though.
        if args:
            attrs: Dict[str, object] = dict(zip(self.__slots__, args))
            attrs.update(kwargs)
            cls = type(self).__name__
            attrs = {
                (f'_{cls}{name}' if name.startswith('__') else name): attr
                for name, attr in attrs.items()
            }
        else:
            attrs = kwargs
        for attr, value in attrs.items():
            object.__setattr__(self, attr, value)

    def __setattr__(self, name: str, value: object) -> None:
        raise AttributeError(f"{type(self)} is immutable")


@API.private
class FinalImmutableMeta(ImmutableMeta):
    def __new__(mcs: Type[FinalImmutableMeta],
                name: str,
                bases: Tuple[type, ...],
                namespace: Dict[str, object]
                ) -> FinalImmutableMeta:

        for b in bases:
            if isinstance(b, FinalImmutableMeta) and b.__module__ != __name__:
                raise TypeError(f"Type '{b.__name__}' cannot be inherited by {name}.")

        return cast(FinalImmutableMeta, super().__new__(mcs, name, bases, namespace))


@API.private
class FinalImmutable(Immutable, metaclass=FinalImmutableMeta):
    __slots__ = ()



import functools
import inspect
from typing import Awaitable, Callable, cast, Dict, Hashable, Sequence

from . import API
from .state import current_container
from .utils import FinalImmutable
from ..core.container import _DEFAULT_SENTINEL, Container
from ..core.exceptions import DependencyNotFoundError

compiled = False


@API.private
class Injection(FinalImmutable):
    """
    Maps an argument name to its dependency and if the injection is required,
    which is equivalent to no default argument.
    """
    __slots__ = ('arg_name', 'required', 'dependency', 'default_value')
    arg_name: str
    required: bool
    dependency: Hashable
    default_value: object

    def __init__(self,
                 *,
                 arg_name: str,
                 required: bool,
                 optional: bool,
                 dependency: Hashable
                 ) -> None:
        super().__init__(
            arg_name=arg_name,
            required=required,
            dependency=dependency,
            default_value=None if optional else _DEFAULT_SENTINEL
        )


@API.private
class InjectionBlueprint(FinalImmutable):
    """
    Stores all the injections for a function.
    """
    __slots__ = ('injections',)
    injections: Sequence[Injection]

    def is_empty(self) -> bool:
        return all(injection.dependency is None for injection in self.injections)


@API.private
def build_wrapper(blueprint: InjectionBlueprint,
                  wrapped: Callable[..., object],
                  skip_self: bool = False) -> Callable[..., object]:
    if inspect.iscoroutinefunction(wrapped):
        return AsyncInjectedWrapper(blueprint,
                                    cast(Callable[..., Awaitable[object]], wrapped),
                                    skip_self)
    return SyncInjectedWrapper(blueprint, wrapped, skip_self)


@API.private
def get_wrapper_injections(wrapper: Callable[..., object]) -> Dict[str, Hashable]:
    if not isinstance(wrapper, InjectedWrapper):
        raise TypeError(f"Argument must be an {InjectedWrapper}")

    if isinstance(wrapper, SyncInjectedWrapper):
        prefix = f"_{SyncInjectedWrapper.__name__}"
    else:
        prefix = f"_{AsyncInjectedWrapper.__name__}"
    blueprint: InjectionBlueprint = getattr(wrapper,
                                            f"{prefix}__blueprint")

    return {inj.arg_name: inj.dependency for inj in blueprint.injections if
            inj.dependency is not None}


@API.private
def is_wrapper(x: object) -> bool:
    return isinstance(x, InjectedWrapper)


@API.private
def get_wrapped(x: object) -> object:
    assert isinstance(x, InjectedWrapper)
    return x.__wrapped__


@API.private
class InjectedWrapper:
    __wrapped__: object


@API.private
class SyncInjectedWrapper(InjectedWrapper):
    """
    Wrapper which injects all the dependencies not supplied in the passed
    arguments. An InjectionBlueprint is used to store the mapping of the
    arguments to their dependency if any and if the injection is required.
    """

    def __init__(self,
                 blueprint: InjectionBlueprint,
                 wrapped: Callable[..., object],
                 skip_self: bool = False):
        """
        Args:
            blueprint: Injection blueprint for the underlying function
            wrapped:  real function to be called
            skip_self:  whether the first argument must be skipped. Used internally
        """
        self.__blueprint = blueprint
        self.__wrapped__: Callable[..., object] = wrapped
        self.__injection_offset = 1 if skip_self else 0
        functools.wraps(wrapped, updated=())(self)

    def __call__(self, *args: object, **kwargs: object) -> object:
        kwargs = _inject_kwargs(
            current_container(),
            self.__blueprint,
            self.__injection_offset + len(args),
            kwargs
        )
        return self.__wrapped__(*args, **kwargs)

    def __get__(self, instance: object, owner: type) -> object:
        return SyncInjectedBoundWrapper(
            self.__blueprint,
            self.__wrapped__.__get__(instance, owner),  # type: ignore
            instance is not None
        )

    def __getattr__(self, item: str) -> object:
        return getattr(self.__wrapped__, item)


@API.private
class SyncInjectedBoundWrapper(SyncInjectedWrapper):
    """
    Behaves like Python bound methods. Unsure whether this is really necessary
    or not.
    """

    def __get__(self, instance: object, owner: type) -> object:
        return self  # pragma: no cover


@API.private
class AsyncInjectedWrapper(InjectedWrapper):
    """
    Wrapper which injects all the dependencies not supplied in the passed
    arguments. An InjectionBlueprint is used to store the mapping of the
    arguments to their dependency if any and if the injection is required.
    """

    def __init__(self,
                 blueprint: InjectionBlueprint,
                 wrapped: Callable[..., Awaitable[object]],
                 skip_self: bool = False):
        """
        Args:
            blueprint: Injection blueprint for the underlying function
            wrapped:  real function to be called
            skip_self:  whether the first argument must be skipped. Used internally
        """
        self.__blueprint = blueprint
        self.__wrapped__: Callable[..., Awaitable[object]] = wrapped
        self.__injection_offset = 1 if skip_self else 0
        functools.wraps(wrapped, updated=())(self)

    async def __call__(self, *args: object, **kwargs: object) -> object:
        kwargs = _inject_kwargs(
            current_container(),
            self.__blueprint,
            self.__injection_offset + len(args),
            kwargs
        )
        return await self.__wrapped__(*args, **kwargs)

    def __get__(self, instance: object, owner: type) -> object:
        return AsyncInjectedBoundWrapper(
            self.__blueprint,
            self.__wrapped__.__get__(instance, owner),  # type: ignore
            instance is not None
        )

    def __getattr__(self, item: str) -> object:
        return getattr(self.__wrapped__, item)


@API.private
class AsyncInjectedBoundWrapper(AsyncInjectedWrapper):
    """
    Behaves like Python bound methods. Unsure whether this is really necessary
    or not.
    """

    def __get__(self, instance: object, owner: type) -> object:
        return self  # pragma: no cover


@API.private
def _inject_kwargs(container: Container,
                   blueprint: InjectionBlueprint,
                   offset: int,
                   kwargs: Dict[str, object]) -> Dict[str, object]:
    """
    Does the actual injection of the dependencies. Used by InjectedCallableWrapper.
    """
    dirty_kwargs = False
    for injection in blueprint.injections[offset:]:
        if injection.dependency is not None and injection.arg_name not in kwargs:
            try:
                arg = container.get(injection.dependency,
                                    injection.default_value)
                if not dirty_kwargs:
                    kwargs = kwargs.copy()
                    dirty_kwargs = True
                kwargs[injection.arg_name] = arg
            except DependencyNotFoundError:
                if injection.required:
                    raise

    return kwargs



import pytest

from antidote import Scope, world
from antidote._providers.lazy import Lazy, LazyProvider
from antidote.core import Container, DependencyValue


@pytest.fixture()
def lazy_provider():
    with world.test.empty():
        world.provider(LazyProvider)
        yield world.get(LazyProvider)


class Dummy(Lazy):
    def __init__(self, value, singleton=False):
        self.value = value
        self.scope = Scope.singleton() if singleton else None

    def __antidote_provide__(self, container: Container) -> DependencyValue:
        return DependencyValue(container.get(self.value), scope=self.scope)


def test_lazy():
    with world.test.empty():
        x = object()
        world.test.singleton(x, object())
        lazy_provider = LazyProvider()

        assert world.test.maybe_provide_from(lazy_provider,
                                             Dummy(x)).unwrapped is world.get(x)
        for s in [True, False]:
            assert world.test.maybe_provide_from(
                lazy_provider,
                Dummy(x, singleton=s)).is_singleton() is s


def test_exists():
    lazy = LazyProvider()
    assert not lazy.exists(object())
    assert lazy.exists(Dummy('x'))


def test_copy(lazy_provider: LazyProvider):
    assert isinstance(lazy_provider.clone(True), LazyProvider)
    assert isinstance(lazy_provider.clone(False), LazyProvider)


def test_debug():
    with pytest.warns(UserWarning, match="(?i).*debug.*"):
        world.debug(Dummy(1))



from typing import Any, Callable, Hashable, Optional

import pytest

from antidote import Scope, Service, world
from antidote.core import DependencyValue
from antidote.exceptions import DependencyNotFoundError
from ..core.utils import DummyProvider

sentinel = object()


@pytest.mark.parametrize('override', [
    pytest.param(lambda d, v: world.test.override.singleton(d, v), id='singleton'),
    pytest.param(lambda d, v: world.test.override.singleton({d: v}), id='singletons')
])
def test_singleton(override: Callable[[Any, Any], Any]):
    with world.test.empty():
        with world.test.clone():
            override('test', 'a')
            assert world.get("test") == 'a'

        world.test.singleton('test', sentinel)

        with world.test.clone(keep_singletons=True):
            assert world.get("test") is sentinel
            override('test', 'a')
            assert world.get("test") == 'a'

    with world.test.clone():
        override("test", 'a')
        assert world.get("test") == 'a'

        # double override works
        override("test", 'b')
        assert world.get("test") == 'b'


def test_factory():
    # overrides normal provider
    with world.test.empty():
        world.provider(DummyProvider)
        p = world.get[DummyProvider]()
        p.data = {'test': sentinel}
        assert world.get('test') is sentinel

        with world.test.clone():
            @world.test.override.factory('test')
            def f():
                return 'a'

            assert world.get('test') == 'a'

            @world.test.override.factory('test')
            def g():
                return 'b'

            assert world.get('test') == 'b'

    # overrides singletons
    with world.test.empty():
        world.test.singleton({'test': sentinel, 'test2': sentinel})
        assert world.get('test') is sentinel
        assert world.get('test2') is sentinel

        with world.test.clone(keep_singletons=True):
            @world.test.override.factory('test')
            def f1():
                return 'a'

            assert world.get('test') == 'a'
            assert world.get('test2') is sentinel

    # default => singleton
    with world.test.clone():
        @world.test.override.factory('test')
        def f2():
            return object()

        assert world.get('test') is world.get('test')

    # non singleton
    with world.test.clone():
        @world.test.override.factory('test', singleton=False)
        def f3():
            return object()

        assert world.get('test') is not world.get('test')

    # singleton
    with world.test.clone():
        @world.test.override.factory('test', singleton=True)
        def f4():
            return object()

        assert world.get('test') is world.get('test')


def test_factory_return_type():
    class X(Service):
        pass

    with world.test.clone():
        @world.test.override.factory()
        def build() -> X:
            return "hello X"

        assert world.get(X) == "hello X"

        with pytest.raises(ValueError, match=".*dependency.*return type hint.*"):
            @world.test.override.factory()
            def build2():
                return "hello X"


def test_provider():
    # overrides normal provider
    with world.test.empty():
        world.provider(DummyProvider)
        p = world.get[DummyProvider]()
        p.data = {'test': sentinel, 'test2': sentinel}
        assert world.get('test') is sentinel
        assert world.get('test2') is sentinel

        with world.test.clone():
            @world.test.override.provider()
            def f(dependency: Hashable) -> Optional[DependencyValue]:
                if dependency == 'test':
                    return DependencyValue('a')

            assert world.get('test') == 'a'
            assert world.get('test2') is sentinel

    # overrides singletons
    with world.test.empty():
        world.test.singleton({'test': sentinel, 'test2': sentinel})
        assert world.get('test') is sentinel
        assert world.get('test2') is sentinel

        with world.test.clone(keep_singletons=True):
            @world.test.override.provider()
            def f2(dependency: Hashable) -> Optional[DependencyValue]:
                if dependency == 'test':
                    return DependencyValue('a')

            assert world.get('test') == 'a'
            assert world.get('test2') is sentinel

    # non singleton dependency
    with world.test.clone():

        @world.test.override.provider()
        def f3(dependency: Hashable) -> Optional[DependencyValue]:
            if dependency == 'test':
                return DependencyValue(object())

        assert world.get('test') is not world.get('test')

    # singleton dependency
    with world.test.clone():

        @world.test.override.provider()
        def f4(dependency: Hashable) -> Optional[DependencyValue]:
            if dependency == 'test':
                return DependencyValue(object(), scope=Scope.singleton())

        assert world.get('test') is world.get('test')


def test_scope_support():
    dummy_scope = world.scopes.new(name='dummy')

    class X(Service):
        __antidote__ = Service.Conf(scope=dummy_scope)

    with world.test.clone():
        x = world.get(X)
        assert x is world.get(X)
        world.scopes.reset(dummy_scope)
        assert world.get(X) is not x

        @world.test.override.factory(1, scope=dummy_scope)
        def build():
            return object()

        x = world.get(1)
        assert x is world.get(1)
        world.scopes.reset(dummy_scope)
        assert world.get(1) is not x

        @world.test.override.provider()
        def provide(dependency):
            if dependency == 2:
                return DependencyValue(object(), scope=dummy_scope)

        x = world.get(2)
        assert x is world.get(2)
        world.scopes.reset(dummy_scope)
        assert world.get(2) is not x


def test_deep_clone():
    with world.test.empty():
        world.test.singleton("test", sentinel)

        with world.test.clone(keep_singletons=True):
            with world.test.clone(keep_singletons=True):
                assert world.get("test") is sentinel

            with world.test.clone(keep_singletons=False):
                with pytest.raises(DependencyNotFoundError):
                    world.get("test")

        with world.test.clone():
            world.test.override.singleton("test", 'a')
            with world.test.clone(keep_singletons=True):
                assert world.get("test") == 'a'

            with world.test.clone(keep_singletons=False):
                with pytest.raises(DependencyNotFoundError):
                    world.get("test")


def test_debug():
    with world.test.new():
        world.test.singleton("original", 2)

        with world.test.clone(keep_singletons=True):
            from antidote._internal import state

            world.test.override.singleton("s", 1)

            @world.test.override.factory('f')
            def f():
                return 2

            c = state.current_container()
            assert "Override" in c.debug("s").info
            assert "Singleton" in c.debug("s").info

            assert "Override" in c.debug("f").info
            assert "Factory" in c.debug("f").info

            assert "Singleton" in c.debug("original").info


def test_invalid():
    with world.test.clone():
        with pytest.raises(TypeError):
            world.test.override.singleton(1)

        with pytest.raises(TypeError):
            world.test.override.singleton(dict(), 1)

        with pytest.raises(TypeError):
            @world.test.override.factory('test', singleton=object())
            def f():
                pass

        with pytest.raises(TypeError):
            world.test.override.factory('test', singleton=True)(object())

        with pytest.raises(TypeError):
            world.test.override.provider()(object())



from __future__ import annotations

from typing import Any, Callable, Optional, Type, TYPE_CHECKING, TypeVar, Union

from .._internal import API
from .._internal.utils import FinalImmutable

if TYPE_CHECKING:
    from .typing import CallableClass, Source

T = TypeVar('T')


@API.private
class Marker:
    pass


@API.private  # See @inject decorator for usage.
class InjectMeMarker(Marker, FinalImmutable):
    __slots__ = ('source',)
    source: Optional[Union[Source[Any], Callable[..., Any], Type[CallableClass[Any]]]]

    def __init__(self,
                 *,
                 source: Optional[Union[
                     Source[Any],
                     Callable[..., Any],
                     Type[CallableClass[Any]]
                 ]]
                 ) -> None:
        super().__init__(source)



from contextlib import contextmanager
from typing import Hashable, Optional

import pytest

from antidote import world
from antidote.core import (Container, DependencyDebug, DependencyValue, does_not_freeze, Provider,
                           StatelessProvider)
from antidote.core.exceptions import DuplicateDependencyError
from antidote.exceptions import FrozenWorldError


@contextmanager
def does_not_raise():
    yield


def test_freeze_world():
    class DummyProvider(Provider):
        def provide(self, dependency: Hashable, container: Container
                    ) -> Optional[DependencyValue]:
            return None

        def clone(self, keep_singletons_cache: bool) -> 'DummyProvider':
            return self

        def register(self):
            return "register"

        @does_not_freeze
        def method(self):
            return "method"

        @staticmethod
        def static():
            return "static"

        @classmethod
        def klass(cls):
            return "klass"

    provider = DummyProvider()
    assert provider.register() == "register"
    assert provider.method() == "method"
    assert provider.static() == "static"
    assert provider.klass() == "klass"

    with world.test.empty():
        world.provider(DummyProvider)
        provider = world.get[DummyProvider]()
        assert provider.register() == "register"
        assert provider.method() == "method"
        assert provider.static() == "static"
        assert provider.klass() == "klass"

        world.freeze()
        assert provider.method() == "method"
        assert provider.static() == "static"
        assert provider.klass() == "klass"
        provider.clone(False)
        provider.provide(None, None)
        with pytest.raises(FrozenWorldError):
            provider.register()


def test_stateless():
    class DummyProvider(StatelessProvider):
        def provide(self, dependency: Hashable, container: Container
                    ) -> Optional[DependencyValue]:
            return None

    p = DummyProvider()
    assert p.clone(True) is not p
    assert isinstance(p.clone(False), DummyProvider)


def test_no_default_implementation():
    class Dummy(Provider):
        pass

    with pytest.raises(NotImplementedError):
        Dummy().maybe_provide(object(), object())

    with pytest.raises(NotImplementedError):
        Dummy().exists(object())

    with pytest.raises(NotImplementedError):
        Dummy().clone(False)

    with pytest.raises(NotImplementedError):
        Dummy().provide(object(), object())

    with pytest.raises(NotImplementedError):
        Dummy().exists(False)


def test_debug():
    x = object()

    class Dummy(Provider):
        def exists(self, dependency: Hashable) -> bool:
            return dependency is x

    Dummy().maybe_debug(object())  # should no fail
    with pytest.warns(UserWarning, match="(?i).*debug.*"):
        Dummy().maybe_debug(x)


def test_provide():
    x = object()

    class Dummy(Provider):
        def exists(self, dependency: Hashable) -> bool:
            return dependency is x

        def provide(self, dependency: Hashable,
                    container: Container) -> DependencyValue:
            assert dependency is x
            return DependencyValue(None)

    dummy = Dummy()
    assert world.test.maybe_provide_from(dummy, 1) is None
    assert world.test.maybe_provide_from(dummy, x) is not None


def test_container_lock():
    from ..test_thread_safety import ThreadSafetyTest

    with world.test.empty():
        failures = []

        @world.provider
        class A(Provider):
            def exists(self, dependency: Hashable) -> bool:
                return dependency == 'a'

            def provide(self, dependency: Hashable,
                        container: Container) -> DependencyValue:
                ThreadSafetyTest.check_locked(failures)
                return DependencyValue('a')

            def change_state(self):
                with self._container_lock():
                    ThreadSafetyTest.check_locked(failures)

        @world.provider
        class B(Provider):
            def exists(self, dependency: Hashable) -> bool:
                return dependency == 'b'

            def provide(self, dependency: Hashable,
                        container: Container) -> DependencyValue:
                ThreadSafetyTest.check_locked(failures)
                return DependencyValue('b')

            def change_state(self):
                with self._container_lock():
                    ThreadSafetyTest.check_locked(failures)

        a = world.get[A]()
        b = world.get[B]()
        actions = [
            lambda: a.change_state(),
            lambda: world.get('a'),
            lambda: b.change_state(),
            lambda: world.get('b')
        ]

        def worker():
            import random
            for f in random.choices(actions, k=20):
                f()

        ThreadSafetyTest.run(worker, n_threads=5)
        assert not failures

    with world.test.empty():
        class C(Provider):
            def change_state(self):
                with self._container_lock():
                    pass

        # Should not raise any error if not bound to any container yet
        C().change_state()


def test_assert_not_duplicate():
    x = object()

    class A(Provider[Hashable]):
        def __init__(self):
            super().__init__()
            self.registered = dict()

        def exists(self, dependency: Hashable) -> bool:
            return dependency in self.registered

        def provide(self,
                    dependency: Hashable,
                    container: Container) -> DependencyValue:
            return DependencyValue(self.registered[dependency])

        def add(self, dependency: Hashable, value: object):
            self._assert_not_duplicate(dependency)
            self.registered[dependency] = value

    with world.test.empty():
        a = A()
        a.add(x, 1)

        with pytest.raises(DuplicateDependencyError):
            with pytest.warns(UserWarning, match="(?i).*debug.*"):
                a.add(x, 1)

    with world.test.empty():
        world.provider(A)

        @world.provider
        class B(Provider[Hashable]):
            def __init__(self):
                super().__init__()
                self.registered = dict()

            def debug(self, dependency: Hashable) -> DependencyDebug:
                return DependencyDebug("DebugInfo")

            def exists(self, dependency: Hashable) -> bool:
                return dependency in self.registered

            def provide(self,
                        dependency: Hashable,
                        container: Container) -> DependencyValue:
                return DependencyValue(self.registered[dependency])

            def add(self, dependency: Hashable, value: object):
                self._assert_not_duplicate(dependency)
                self.registered[dependency] = value

        world.get[B]().add(x, 1)

        with pytest.raises(DuplicateDependencyError, match=".*DebugInfo.*"):
            world.get[A]().add(x, 1)



# pyright: reportInvalidTypeVarUse=false
from __future__ import annotations

from typing import Any, Generic, Type, TypeVar

from typing_extensions import Protocol, runtime_checkable

from .._internal import API

Tct = TypeVar('Tct', contravariant=True)
Tco = TypeVar('Tco', covariant=True)
T = TypeVar('T')


@API.private
@runtime_checkable
class Source(Protocol[Tct]):
    def __antidote_dependency__(self, dependency: Type[Tct]) -> object:
        """
        The input type is not guaranteed. You MUST check the type. Raise an error if this
        source cannot provide the specified target.
        """


@API.private
class CallableClass(Protocol[Tco]):
    def __call__(self, *args: Any, **kwargs: Any) -> Tco:
        ...  # pragma: no cover


@API.private
class Dependency(Generic[T]):
    pass



import itertools
from typing import Any, Callable, Optional, Tuple, Type, Union

import pytest
from typing_extensions import Protocol

from antidote import Constants, Factory, Provide, Service, service, world
from antidote.core import Wiring


class Interface:
    pass


class FactoryOutput:
    pass


class FactoryOutput2:
    pass


class A:
    pass


class B:
    pass


@pytest.fixture(autouse=True)
def new_world():
    with world.test.new():
        world.test.singleton({A: A(),
                              B: B(),
                              'a': object(),
                              'b': object(),
                              'x': object(),
                              'y': object(),
                              'z': object()})
        yield


class DummyProtocol(Protocol):
    a: A
    b: B

    def __init__(self, a: A = None, b: B = None):
        pass

    def method_AB(self, a: A = None, b: B = None) -> Tuple[A, B]:
        pass

    def method_ab(self, a=None, b=None) -> Tuple[Any, Any]:
        pass

    def method_xyz(self, x=None, y=None, z=None) -> Tuple[Any, Any, Any]:
        pass


DEFAULT_WIRING = object()


def builder(cls_or_decorator: Union[type, Callable[..., Any]],
            wiring_kind: str = 'Wiring',
            subclass: bool = False):
    meta_kwargs = dict(abstract=True) if subclass else dict()
    if isinstance(cls_or_decorator, type):
        cls = cls_or_decorator

        def decorator(wiring=None):
            return lambda x: x
    else:
        cls = object
        decorator = cls_or_decorator

    def build(wiring: Wiring = None):
        decorate = (decorator(wiring=wiring)
                    if wiring is not DEFAULT_WIRING else
                    decorator())

        @decorate
        class Dummy(cls, **meta_kwargs):
            if wiring is not DEFAULT_WIRING and cls is not object:
                if wiring is not None:
                    if wiring_kind == 'Wiring':
                        __antidote__ = cls.Conf(wiring=wiring)
                    else:
                        __antidote__ = cls.Conf().with_wiring(**{
                            attr: getattr(wiring, attr)
                            for attr in Wiring.__slots__
                        })
                else:
                    __antidote__ = cls.Conf(wiring=None)

            def __init__(self, a: Provide[A] = None, b: Provide[B] = None):
                super().__init__()
                self.a = a
                self.b = b

            def method_AB(self, a: A = None, b: B = None) -> Tuple[A, B]:
                return a, b

            def method_ab(self, a=None, b=None) -> Tuple[Any, Any]:
                return a, b

            def method_xyz(self, x=None, y=None, z=None) -> Tuple[Any, Any, Any]:
                return x, y, z

            def __call__(self) -> FactoryOutput:  # for Factory
                pass

            def get(self):  # for Constants
                pass

        if subclass:
            class SubDummy(Dummy):
                def __call__(self) -> FactoryOutput2:  # for Factory
                    pass

            return SubDummy
        else:
            return Dummy

    return build


@pytest.fixture(params=[
    pytest.param((builder, service), id="@service"),
    *[
        pytest.param((builder, c, w), id=f"{c.__name__} - {w}")
        for (c, w) in itertools.product([Factory,
                                         Service,
                                         Constants],
                                        ['with_wiring', 'Wiring'])
    ]
])
def class_builder(request):
    f, *args = request.param
    return f(*args)


@pytest.fixture(params=[
    pytest.param((c, w), id=f"{c.__name__} - w")
    for (c, w) in itertools.product([Factory,
                                     Service,
                                     Constants],
                                    ['with_wiring', 'Wiring'])
])
def subclass_builder(request):
    (cls, wiring_kind) = request.param
    return builder(cls, wiring_kind, subclass=True)


F = Callable[[Optional[Wiring]], Type[DummyProtocol]]


def test_default(class_builder: F):
    dummy = class_builder(DEFAULT_WIRING)()
    assert dummy.a is world.get(A)
    assert dummy.b is world.get(B)

    (a, b) = dummy.method_AB()
    assert a is None
    assert b is None


def test_no_wiring(class_builder: F):
    dummy = class_builder(None)()
    assert dummy.a is None
    assert dummy.b is None


def test_methods(class_builder: F):
    dummy = class_builder(Wiring(methods=('__init__',)))()
    assert dummy.a is world.get(A)
    assert dummy.b is world.get(B)

    (a, b) = dummy.method_AB()
    assert a is None
    assert b is None

    dummy = class_builder(Wiring(methods=('method_AB',),
                                 auto_provide=True))()
    assert dummy.a is None
    assert dummy.b is None

    (a, b) = dummy.method_AB()
    assert a is world.get(A)
    assert b is world.get(B)


def test_auto_provide(class_builder: F):
    # Uses type hints by default
    dummy = class_builder(Wiring(methods=('method_AB',)))()
    (a, b) = dummy.method_AB()
    assert a is None
    assert b is None

    dummy = class_builder(Wiring(methods=('method_AB',), auto_provide=True))()
    (a, b) = dummy.method_AB()
    assert a is world.get(A)
    assert b is world.get(B)

    dummy = class_builder(Wiring(methods=('method_AB',), auto_provide=False))()
    (a, b) = dummy.method_AB()
    assert a is None
    assert b is None

    dummy = class_builder(Wiring(methods=('method_AB',), auto_provide=[A]))()
    (a, b) = dummy.method_AB()
    assert a is world.get(A)
    assert b is None

    dummy = class_builder(Wiring(methods=('method_AB',),
                                 auto_provide=lambda cls: True))()
    (a, b) = dummy.method_AB()
    assert a is world.get(A)
    assert b is world.get(B)

    dummy = class_builder(Wiring(methods=('method_AB',),
                                 auto_provide=lambda cls: issubclass(cls, A)))()
    (a, b) = dummy.method_AB()
    assert a is world.get(A)
    assert b is None

    dummy = class_builder(Wiring(methods=('method_AB',),
                                 auto_provide=lambda cls: False))()
    (a, b) = dummy.method_AB()
    assert a is None
    assert b is None


def test_dependencies_dict(class_builder: F):
    dummy = class_builder(Wiring(dependencies=dict(),
                                 auto_provide=True))()
    assert dummy.a is world.get(A)
    assert dummy.b is world.get(B)
    (a, b) = dummy.method_AB()
    assert a is world.get(A)
    assert b is world.get(B)
    (a, b) = dummy.method_ab()
    assert a is None
    assert b is None

    dummy = class_builder(Wiring(dependencies=dict(a='x', b='y'),
                                 auto_provide=True))()
    assert dummy.a is world.get(A)
    assert dummy.b is world.get(B)
    (a, b) = dummy.method_AB()
    assert a is world.get('x')
    assert b is world.get('y')
    (a, b) = dummy.method_ab()
    assert a is world.get('x')
    assert b is world.get('y')

    dummy = class_builder(Wiring(dependencies=dict(b='y'),
                                 auto_provide=True))()
    assert dummy.a is world.get(A)
    assert dummy.b is world.get(B)
    (a, b) = dummy.method_AB()
    assert a is world.get(A)
    assert b is world.get('y')
    (a, b) = dummy.method_ab()
    assert a is None
    assert b is world.get('y')


def test_dependencies_seq(class_builder: F):
    dummy = class_builder(Wiring(dependencies=[],
                                 auto_provide=True))()
    assert dummy.a is world.get(A)
    assert dummy.b is world.get(B)
    (a, b) = dummy.method_AB()
    assert a is world.get(A)
    assert b is world.get(B)
    (a, b) = dummy.method_ab()
    assert a is None
    assert b is None

    dummy = class_builder(Wiring(dependencies=[None, None],
                                 auto_provide=True))()
    assert dummy.a is world.get(A)
    assert dummy.b is world.get(B)
    (a, b) = dummy.method_AB()
    assert a is world.get(A)
    assert b is world.get(B)
    (a, b) = dummy.method_ab()
    assert a is None
    assert b is None

    dummy = class_builder(Wiring(dependencies=['x', 'y'],
                                 auto_provide=True))()
    assert dummy.a is world.get(A)
    assert dummy.b is world.get(B)
    (a, b) = dummy.method_AB()
    assert a is world.get('x')
    assert b is world.get('y')
    (a, b) = dummy.method_ab()
    assert a is world.get('x')
    assert b is world.get('y')

    dummy = class_builder(Wiring(dependencies=[None, 'y'],
                                 auto_provide=True))()
    assert dummy.a is world.get(A)
    assert dummy.b is world.get(B)
    (a, b) = dummy.method_AB()
    assert a is world.get(A)
    assert b is world.get('y')
    (a, b) = dummy.method_ab()
    assert a is None
    assert b is world.get('y')

    dummy = class_builder(Wiring(dependencies=['x', None],
                                 auto_provide=True))()
    assert dummy.a is world.get(A)
    assert dummy.b is world.get(B)
    (a, b) = dummy.method_AB()
    assert a is world.get('x')
    assert b is world.get(B)
    (a, b) = dummy.method_ab()
    assert a is world.get('x')
    assert b is None


def test_dependencies_callable(class_builder: F):
    dummy = class_builder(Wiring(dependencies=lambda arg: None,
                                 auto_provide=True))()
    assert dummy.a is world.get(A)
    assert dummy.b is world.get(B)
    (a, b) = dummy.method_AB()
    assert a is world.get(A)
    assert b is world.get(B)
    (a, b) = dummy.method_ab()
    assert a is None
    assert b is None

    dummy = class_builder(Wiring(dependencies=lambda arg: 'x' if arg.name == 'a' else 'y',
                                 auto_provide=True))()
    assert dummy.a is world.get(A)
    assert dummy.b is world.get(B)
    (a, b) = dummy.method_AB()
    assert a is world.get('x')
    assert b is world.get('y')
    (a, b) = dummy.method_ab()
    assert a is world.get('x')
    assert b is world.get('y')

    dummy = class_builder(Wiring(
        dependencies=lambda arg: 'x' if arg.name == 'a' else None,
        auto_provide=True))()
    assert dummy.a is world.get(A)
    assert dummy.b is world.get(B)
    (a, b) = dummy.method_AB()
    assert a is world.get('x')
    assert b is world.get(B)
    (a, b) = dummy.method_ab()
    assert a is world.get('x')
    assert b is None


def test_distinct_arguments(class_builder: F):
    # Having more arguments in dependencies seq is not an issue for methods having less.
    dummy = class_builder(Wiring(methods=('method_AB', 'method_xyz'),
                                 dependencies=['x', None, 'z'],
                                 auto_provide=True))()
    (a, b) = dummy.method_AB()
    assert a is world.get('x')
    assert b is world.get(B)
    (x, y, z) = dummy.method_xyz()
    assert x is world.get('x')
    assert y is None
    assert z is world.get('z')

    # Unknown argument in the dependencies dict won't raise an error.
    dummy = class_builder(Wiring(methods=('method_AB', 'method_xyz'),
                                 dependencies=dict(b='b', y='y'),
                                 auto_provide=True))()
    (a, b) = dummy.method_AB()
    assert a is world.get(A)
    assert b is world.get('b')
    (x, y, z) = dummy.method_xyz()
    assert x is None
    assert y is world.get('y')
    assert z is None

    # type_hints
    dummy = class_builder(Wiring(methods=('method_AB', 'method_xyz'),
                                 auto_provide=[A]))()
    (a, b) = dummy.method_AB()
    assert a is world.get(A)
    assert b is None
    (x, y, z) = dummy.method_xyz()
    assert x is None
    assert y is None
    assert z is None



from contextlib import contextmanager
from typing import Hashable

# @formatter:off
cimport cython
from cpython.mem cimport PyMem_Free, PyMem_Malloc, PyMem_Realloc
from cpython.object cimport PyObject
# @formatter:on

cdef extern from "Python.h":
    Py_hash_t PyObject_Hash(PyObject *o)


@cython.final
cdef class DependencyStack:
    """
    Cython implementation of the DependencyStack with the same Python API.

    It relies on an array of hashes which is faster than a set for a low depth.
    """

    def __init__(self):
        cdef:
            size_t capacity = 32

        self._trace = <PyObject**> PyMem_Malloc(capacity * sizeof(PyObject *))
        self._hashes = <Py_hash_t *> PyMem_Malloc(capacity * sizeof(Py_hash_t))
        self._depth = 0
        self._capacity = capacity

    def __dealloc__(self):
        PyMem_Free(self._trace)
        PyMem_Free(self._hashes)

    @contextmanager
    def instantiating(self, dependency: Hashable):
        if 0 != self.push(<PyObject *> dependency):
            raise self.reset_with_error(<PyObject *> dependency)
        try:
            yield
        finally:
            self.pop()

    @property
    def depth(self):
        return self._depth

    def to_list(self):
        cdef:
            list l = []
            PyObject *t

        for t in self._trace[:self._depth]:
            l.append(<object> t)

        return l

    cdef Exception reset_with_error(self, PyObject *dependency):
        cdef:
            list l = []
            PyObject *t

        for t in self._trace[:self._depth]:
            l.append(<object> t)
        l.append(<object> dependency)

        from ..core.exceptions import DependencyCycleError
        return DependencyCycleError(l)

    cdef bint push(self, PyObject *dependency):
        """
        Args:
            dependency: supposed to be hashable as the core tries to 
                retrieves from a dictionary first. 

        Returns:
            0 OK
            1 if present in the stack
        """
        cdef:
            size_t depth = self._depth
            PyObject** traces = self._trace
            Py_hash_t *hashes = self._hashes
            Py_hash_t *h
            Py_hash_t dependency_hash = PyObject_Hash(dependency)
            int i = 0
            PyObject *t

        for h in hashes[:depth]:
            if h[0] == dependency_hash:
                t = traces[i]
                if dependency == t or <object> dependency == <object> t:
                    return 1
            i += 1

        if depth == self._capacity:
            self._capacity *= 2
            PyMem_Realloc(hashes, self._capacity * sizeof(Py_hash_t))
            PyMem_Realloc(traces, self._capacity * sizeof(PyObject *))
        hashes[depth] = dependency_hash
        traces[depth] = dependency

        self._depth += 1

        return 0

    cdef void pop(self):
        """
        Latest elements of the stack is removed.
        """
        self._depth -= 1



from .annotations import From, FromArg, Get, Inject, Provide
from .container import Container, DependencyValue, Scope
from .injection import Arg, DEPENDENCIES_TYPE, inject
from .provider import does_not_freeze, Provider, StatelessProvider
from .typing import Source, Dependency
from .utils import DependencyDebug
from .wiring import wire, Wiring, WithWiringMixin

__all__ = ['Provide', 'Inject', 'Get', 'From', 'FromArg',
           'Container', 'DependencyValue', 'Scope', 'inject', 'Arg',
           'DEPENDENCIES_TYPE', 'does_not_freeze', 'Provider', 'StatelessProvider',
           'DependencyDebug', 'wire', 'Wiring', 'WithWiringMixin', 'Source', 'Dependency']



from contextlib import contextmanager
from typing import (Any, Callable, cast, Dict, get_type_hints, Hashable, Iterator, Optional,
                    overload,
                    TypeVar, Union)

from ..._internal import API, state
from ..._providers.world_test import WorldTestProvider
from ...core import inject
from ...core.container import DependencyValue, RawContainer, RawProvider, Scope
from ...core.exceptions import DependencyNotFoundError
from ...utils import validated_scope


@API.public
@contextmanager
def clone(*,
          keep_singletons: bool = False,
          keep_scopes: bool = False
          ) -> Iterator[None]:
    """
    Clone the current container, namely scopes and providers. Existing singletons and
    scopes depdency values can also be propagated. Its primary use is to test existing
    dependencies. Hence the world will already be frozen but overridding dependencies
    will be possible.

    .. doctest:: world_test_clone

        >>> from antidote import world, Service
        >>> class MyService(Service):
        ...     pass
        >>> # Declared dependencies are available
        ... with world.test.clone():
        ...     world.get[MyService]()
        <MyService object ...>
        >>> s = world.get[MyService]()
        >>> # But by default singletons are NOT kept
        ... with world.test.clone():
        ...     world.get[MyService]() is s
        False
        >>> # which is of course configurable
        ... with world.test.clone(keep_singletons=True):
        ...     world.get[MyService]() is s
        True
        >>> # You'll keep the same objects though, so be careful !
        ... with world.test.clone(keep_singletons=True):
        ...     world.get[MyService]().new_attr = "hello"
        >>> world.get[MyService]().new_attr
        'hello'
        >>> # You cannot define new dependencies
        ... with world.test.clone():
        ...     class NewService(Service):
        ...         pass
        Traceback (most recent call last):
        ...
        FrozenWorldError
        >>> # but they can be overridden
        ... with world.test.clone():
        ...     # dependencies can also be overridden as many times as necessary
        ...     world.test.override.singleton(MyService, "fake service")
        ...     world.get(MyService)
        'fake service'

    Args:
        keep_singletons: Whether current singletons should be kept in the new world.
            Beware that the same objects will be propagated. Defaults to :py:obj:`False`.
        keep_scopes: Whether current dependency values within the existing scopes should
            be kept in the new world. Like :code:`keep_singletons`, the same objects will
            be propagated. Defaults to :py:obj:`False`.

    Returns:
        context manager within you can make your tests.

    """

    def build(c: RawContainer) -> RawContainer:
        return c.clone(keep_singletons=keep_singletons,
                       keep_scopes=keep_scopes)

    with state.override(build):
        yield


@API.public
@contextmanager
def new() -> Iterator[None]:
    """
    Creates a new container with the same kind of providers and scopes. So it doesn't
    have any dependencies defined, but any scope or provider added will be present.

    .. doctest:: world_test_new

        >>> from antidote import world
        >>> with world.test.new():
        ...     # Current world is empty but has the same providers and scope as the
        ...     # original one.
        ...     # You can define dependencies locally with world.test.singleton or factory
        ...     world.test.singleton("test", 1)
        ...     world.get[int]("test")
        ...
        1
        >>> # Anything done within the context manager stays there.
        ... world.get[int]("test")
        Traceback (most recent call last):
        ...
        DependencyNotFoundError: test

    """

    def build_new_container(existing: RawContainer) -> RawContainer:
        c = RawContainer.with_same_providers_and_scopes(existing)
        try:
            c.get(WorldTestProvider)
        except DependencyNotFoundError:
            c.add_provider(WorldTestProvider)
        return c

    with state.override(build_new_container):
        yield


#######################################
# Utilities usable inside test worlds #
#######################################

__sentinel = object()


@overload
def singleton(dependency: Hashable, value: object) -> None:
    ...  # pragma: no cover


@overload
def singleton(dependency: Dict[Hashable, object]) -> None:
    ...  # pragma: no cover


@API.public
@inject
def singleton(dependency: Union[Dict[object, object], object],
              value: object = __sentinel,
              test_provider: Optional[WorldTestProvider] = inject.me()) -> None:
    """
    Declare one or multiple singleton dependencies with its associated value.

    .. doctest:: world_test_singleton

        >>> from antidote import world
        >>> with world.test.new():  # or empty()
        ...     world.test.singleton("test", 1)
        ...     # or
        ...     world.test.singleton({"dummy": 1})
        ...     world.get[int]("test")
        1

    Args:
        test_provider:
        dependency: Singleton to declare, must be hashable. If a dict is provided, it'll
            be treated as a dictionary of singletons to add.
        value: Associated value for the dependency.

    """
    if test_provider is None:
        raise RuntimeError("Test singletons can only be added inside a test world "
                           "created with world.test.new() or world.test.empty()")
    if value is __sentinel:
        if isinstance(dependency, dict):
            test_provider.add_singletons(cast(Dict[object, object], dependency))
        else:
            raise TypeError("If only a single argument is provided, "
                            "it must be a dictionary of singletons.")
    else:
        if isinstance(dependency, dict):
            raise TypeError("A dictionary cannot be used as a key.")
        test_provider.add_singletons({dependency: value})


F = TypeVar('F', bound=Callable[..., Any])


@API.public
def factory(dependency: Hashable = None,
            *,
            singleton: Optional[bool] = None,
            scope: Optional[Scope] = Scope.sentinel()
            ) -> Callable[[F], F]:
    """
    Declare one or multiple singleton dependencies with its associated value.

    .. doctest:: world_test_factory

        >>> from antidote import world
        >>> class Dummy:
        ...     pass
        >>> with world.test.new():  # or empty()
        ...     @world.test.factory()
        ...     def build_dummy() -> Dummy:
        ...         return Dummy()
        ...     world.get[Dummy]()
        <Dummy ...>
        >>> with world.test.new():
        ...     # You don't need to rely on type hints
        ...     @world.test.factory(Dummy)
        ...     def build_dummy():
        ...         return Dummy()
        ...     world.get[Dummy]()
        <Dummy ...>

    Args:
        dependency: Dependency to override.
        singleton: Whether the returned dependency  is a singleton or not. If yes,
            the factory will be called at most once and the result re-used. Mutually
            exclusive with :code:`scope`. Defaults to :py:obj:`True`.
        scope: Scope of the returned dependency. Mutually exclusive with
            :code:`singleton`. The scope defines if and how long the returned dependency
            will be cached. See :py:class:`~.core.container.Scope`. Defaults to
            :py:meth:`~.core.container.Scope.singleton`.

    """
    scope = validated_scope(scope, singleton, default=Scope.singleton())

    @inject
    def decorate(f: F, test_provider: Optional[WorldTestProvider] = inject.me()) -> F:
        if test_provider is None:
            raise RuntimeError("Test singletons can only be added inside a test world "
                               "created with world.test.new()")
        if not callable(f):
            raise TypeError(f"factory must be a callable, not a {type(f)}")
        if dependency is None:
            output = get_type_hints(f).get('return')
            if output is None:
                raise ValueError("Either the dependency argument or the return type hint "
                                 "of the factory must be specified")
        else:
            output = dependency
        test_provider.add_factory(output, factory=f, scope=scope)
        return f

    return decorate


###############################
# Utilities to test providers #
###############################

@API.public
@contextmanager
def empty() -> Iterator[None]:
    """
    Only used to test providers.

    Creates an empty container. No initial providers are set up. This is useful to test
    one or several providers in isolation. If you're not testing provider, consider use
    :py:func:`.new` instead.
    """

    def build_container(_: RawContainer) -> RawContainer:
        c = RawContainer()
        c.add_provider(WorldTestProvider)
        return c

    with state.override(build_container):
        yield


@API.public
def maybe_provide_from(provider: RawProvider,
                       dependency: Hashable) -> Optional[DependencyValue]:
    """
    Only used to test providers.

    Utility function to test _providers that have not been registered in :py:mod:`.world`.
    The current :py:class:~.core.container.DependencyContainer` will be given as if it
    were registered. This allows to test exactly what has been returned by the provider.

    Args:
        provider: provider from which the dependency should be retrieved.
        dependency: dependency to retrieve.

    Returns:
        dependency instance as returned by the provider.
    """
    if provider.is_registered:
        raise RuntimeError("Method only intended to test provider that have not "
                           "been registered in world.")
    return provider.maybe_provide(dependency, state.current_container())



import pytest

from antidote.exceptions import (DependencyCycleError, DependencyInstantiationError,
                                 DependencyNotFoundError, DuplicateDependencyError,
                                 FrozenWorldError)


class Service:
    pass


class LongRepr:
    def __repr__(self):
        return "\n'test'\n"


@pytest.mark.parametrize('error', [
    pytest.param(DependencyCycleError([Service, 'test', Service]), id='cycle'),
    pytest.param(DependencyCycleError([Service, LongRepr(), Service]), id='cycle-breaks'),
    pytest.param(DependencyInstantiationError(Service, ['test', Service]),
                 id='instantiation'),
    pytest.param(DependencyInstantiationError(Service, [LongRepr(), Service]),
                 id='instantiation-breaks'),
])
def test_stack_error(error):
    for f in [str, repr]:
        assert f"{Service.__module__}.{Service.__name__}" in f(error)
        assert "'test'" in f(error)


def test_dependency_not_found():
    o = object()
    error = DependencyNotFoundError(o)

    for f in [str, repr]:
        assert repr(o) in f(error)


def test_duplicate_dependency_error():
    message = "hello"
    error = DuplicateDependencyError(message)
    assert message in str(error)


def test_frozen_world():
    message = "my message"
    error = FrozenWorldError(message)
    assert message in str(error)
    assert message in repr(error)


def test_duplicate_dependency():
    x = object()
    y = object()
    error = DuplicateDependencyError(x, y)
    for f in [str, repr]:
        assert f(x) in f(error)
        assert f(y) in f(error)

    message = "test"
    error = DuplicateDependencyError(message)
    assert str(message) in str(error)
    assert str(message) in repr(error)



from __future__ import annotations

import threading
from collections import deque
from contextlib import contextmanager
from typing import (Callable, Deque, Dict, Hashable, Iterator, List, Optional, Sequence, Tuple,
                    Type, TYPE_CHECKING)
from weakref import ref, ReferenceType

from typing_extensions import final

from .exceptions import (DependencyCycleError, DependencyInstantiationError,
                         DependencyNotFoundError, DuplicateDependencyError,
                         FrozenWorldError)
from .._internal import API
from .._internal.stack import DependencyStack
from .._internal.utils import FinalImmutable

if TYPE_CHECKING:
    from .utils import DependencyDebug

# PRIVATE
_CONTAINER_REF_ATTR = "_antidote__container_ref"


@API.public
@final
class Scope(FinalImmutable):
    """
    Used to identify a specific scope for dependencies. The scope of a dependency defines
    when the dependency value is valid or not. Or said differently, for how long it is
    valid. A singleton is valid forever, and no scope at all means that a new dependency
    value needs to be retrieved every time.

    Scopes can be create through :py:func:`.world.scopes.new`. The name is only used to
    have a friendly identifier when debugging.

    .. doctest:: core_container_scope

        >>> from antidote import world
        >>> REQUEST_SCOPE = world.scopes.new(name='request')

    To use the newly created scope, use :code:`scope` parameters:

    .. doctest:: core_container_scope

        >>> from antidote import Service
        >>> class Dummy(Service):
        ...     __antidote__ = Service.Conf(scope=REQUEST_SCOPE)

    As :code:`Dummy` has been defined with a custom scope, the dependency value will
    be kep as long as :code:`REQUEST_SCOPE` stays valid. That is to say, until you reset
    it with :py:func:`.world.scopes.reset`:

    .. doctest:: core_container_scope

        >>> dummy = world.get[Dummy]()
        >>> dummy is world.get(Dummy)
        True
        >>> world.scopes.reset(REQUEST_SCOPE)
        >>> dummy is world.get(Dummy)
        False

    .. note::

        You probably noticed that dependencies supporting scopes always offer both
        :code:`singleton` and :code:`scope` arguments. Those are mutually exclusive.
        The reason behind this is simply due to the nature of scopes. Scopes are hard
        to get right ! For example, what should happen with a service in a scope A
        which required a dependency in scope B when B resets ? What if the service kept
        the dependency as an attribute ? There are no good answers for this, hence
        scopes are a pretty advanced feature which can cause inconsistencies. Moreover
        they do have a performance impact. But at the same time, scopes aren't uncommon.
        In a webapp you'll typically need soon enough a request scope. So it shouldn't
        be hard to use it.

        So to have easy to use scopes while keeping them under the radar until you
        actually need them, Antidote exposes both :code:`singleton` and :code:`scope`
        arguments.

    """
    __slots__ = ('name',)
    name: str

    def __repr__(self) -> str:
        return f"Scope(name='{self.name}')"

    @staticmethod
    @API.public
    def singleton() -> Scope:
        """
        Using this scope or specifying :code:`singleton=True` is equivalent.

        Returns:
            Singleton scope (unique object).
        """
        return _SCOPE_SINGLETON

    @staticmethod
    @API.public
    def sentinel() -> Scope:
        """
        For functions having both :code:`singleton` and :code:`scope` argument, validation
        of those is done with :py:func:`~.utils.validated_scope`. To correctly identify
        if the scope was actually set by the user, we're using this sentinel scope as the
        default value.

        Returns:
            Sentinel scope (unique object).
        """
        return _SCOPE_SENTINEL


# Use Scope.singleton() instead.
# API.private
_SCOPE_SINGLETON = Scope('singleton')
_SCOPE_SENTINEL = Scope('__sentinel__')


@API.public
@final
class DependencyValue(FinalImmutable):
    """
    Simple wrapper of the dependency value given by a
    :py:class:`~.provider.Provider`.
    """
    __slots__ = ('unwrapped', 'scope')
    unwrapped: object
    """Actual dependency value."""
    scope: Optional[Scope]
    """Scope of the dependency."""

    def __init__(self,
                 value: object,
                 *,
                 scope: Optional[Scope] = None) -> None:
        """
        Args:
            value: Actual dependency value.
            scope: Scope of the dependency.
        """
        assert scope is not _SCOPE_SENTINEL
        super().__init__(unwrapped=value, scope=scope)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, DependencyValue) \
               and self.unwrapped == other.unwrapped \
               and self.scope is other.scope  # noqa: E126

    def is_singleton(self) -> bool:
        return self.scope is _SCOPE_SINGLETON


# API.private
_DEFAULT_SENTINEL = object()


@API.public
class Container:
    """
    Public interface of the container used by Antidote to handles all
    dependencies. Used in a :py:class:`~.provider.Provider` to access other
    dependencies.
    """

    def get(self, dependency: Hashable, default: object = _DEFAULT_SENTINEL) -> object:
        """
        Retrieves given dependency or raises a
        :py:exc:`~..exceptions.DependencyNotFoundError`.

        Args:
            dependency: Dependency to be retrieved.
            default: Returned if specified and no dependency can be retrieved.

        Returns:
            The dependency instance or the default if specified.
        """
        raise NotImplementedError()  # pragma: no cover

    def provide(self, dependency: Hashable) -> DependencyValue:
        """
        Similar to :py:meth:`~.get` it will retrieve a dependency. However it will
        be wrapped in a :py:class:`~.DependencyValue` if found. If not
        :py:obj:`None` will be returned. This allows to get additional information such
        as whether the dependency is a singleton or not.

        Args:
            dependency: Dependency to be retrieved.

        Returns:
            wrapped dependency instance.

        """
        raise NotImplementedError()  # pragma: no cover


###########
# PRIVATE #
###########

@API.private
class RawProvider:
    """
    Abstract base class for a Provider.

    Use either :py:class:`~.core.provider.Provider` or
    :py:class:`~.core.provider.StatelessProvider` which are easier to implement.

    :meta private:
    """

    def __init__(self) -> None:
        setattr(self, _CONTAINER_REF_ATTR, None)

    def clone(self, keep_singletons_cache: bool) -> RawProvider:
        raise NotImplementedError()  # pragma: no cover

    def exists(self, dependency: Hashable) -> bool:
        raise NotImplementedError()  # pragma: no cover

    def maybe_provide(self, dependency: Hashable, container: Container
                      ) -> Optional[DependencyValue]:
        raise NotImplementedError()  # pragma: no cover

    def maybe_debug(self, dependency: Hashable) -> Optional[DependencyDebug]:
        raise NotImplementedError()  # pragma: no cover

    @API.private
    @final
    @contextmanager
    def _bound_container_locked(self, *, freezing: bool = False) -> Iterator[None]:
        container = self.__bound_container()
        if container is not None:
            with container.locked(freezing=freezing):
                yield
        else:
            yield

    @API.private
    @final
    def _bound_container_raise_if_exists(self, dependency: Hashable) -> None:
        container = self.__bound_container()
        if container is not None:
            container.raise_if_exists(dependency)
        else:
            if self.exists(dependency):
                raise DuplicateDependencyError(
                    f"{dependency} has already been registered in {type(self)}")

    @API.private
    def __bound_container(self) -> Optional[RawContainer]:
        container_ref: 'ReferenceType[RawContainer]' = getattr(self,
                                                               _CONTAINER_REF_ATTR)
        if container_ref is not None:
            container = container_ref()
            assert container is not None, "Bound container does not exist anymore."
            return container
        return None

    # API.private
    @property
    def is_registered(self) -> bool:
        return getattr(self, _CONTAINER_REF_ATTR) is not None


@API.private  # Not meant for direct use. You MUST go through world to manipulate it.
class RawContainer(Container):
    def __init__(self) -> None:
        self._dependency_stack = DependencyStack()
        self._registration_lock = threading.RLock()
        self._instantiation_lock = threading.RLock()

        self.__frozen = False
        self.__singletons: Dict[object, object] = dict()
        self.__scopes: Dict[Scope, Dict[object, object]] = dict()
        self.__providers: List[RawProvider] = list()

    def __repr__(self) -> str:
        return f"{type(self).__name__}(providers={', '.join(map(str, self.__providers))})"

    @staticmethod
    def with_same_providers_and_scopes(original: 'RawContainer') -> RawContainer:
        container = RawContainer()
        for provider in original.providers:
            container.add_provider(type(provider))
        container.__scopes = {scope: dict() for scope in original.__scopes.keys()}
        return container

    @property
    def scopes(self) -> Sequence[Scope]:
        return list(self.__scopes.keys())

    @property
    def providers(self) -> Sequence[RawProvider]:
        return self.__providers.copy()

    @contextmanager
    def locked(self, *, freezing: bool = False) -> Iterator[None]:
        assert isinstance(freezing, bool)
        with self._registration_lock, self._instantiation_lock:
            if freezing and self.__frozen:
                raise FrozenWorldError()
            yield

    def freeze(self) -> None:
        with self._registration_lock:
            if self.__frozen:
                raise FrozenWorldError("Container is already frozen !")
            self.__frozen = True

    def add_provider(self, provider_cls: Type[RawProvider]) -> None:
        with self.locked(freezing=True):
            assert all(provider_cls != type(p) for p in self.__providers)
            provider = provider_cls()
            setattr(provider, _CONTAINER_REF_ATTR, ref(self))
            self.__providers.append(provider)
            self.__singletons[provider_cls] = provider

    def create_scope(self, name: str) -> Scope:
        scope = Scope(name)  # Name is only a helper, not a identifier by itself.
        with self.locked(freezing=True):
            assert all(s.name != name for s in self.__scopes.keys())
            assert len(self.__scopes) < 255  # Consistency with Cython.
            self.__scopes[scope] = dict()
        return scope

    def reset_scope(self, scope: Scope) -> None:
        with self._instantiation_lock:
            self.__scopes[scope].clear()

    def raise_if_exists(self, dependency: Hashable) -> None:
        with self._registration_lock:
            for provider in self.__providers:
                if provider.exists(dependency):
                    debug = provider.maybe_debug(dependency)
                    message = f"{dependency!r} has already been declared " \
                              f"in {type(provider)}"
                    if debug is None:
                        raise DuplicateDependencyError(message)
                    else:
                        raise DuplicateDependencyError(message + f"\n{debug.info}")

    def clone(self,
              *,
              keep_singletons: bool = False,
              keep_scopes: bool = False) -> OverridableRawContainer:
        with self.locked():
            clone = OverridableRawContainer()
            clone.__frozen = True
            if keep_singletons:
                clone.__singletons = self.__singletons.copy()

            clone.__scopes = {
                scope: dependencies.copy() if keep_scopes else dict()
                for scope, dependencies in self.__scopes.items()
            }

            for p in self.__providers:
                p_clone = p.clone(keep_singletons_cache=keep_singletons)
                if p_clone is p \
                        or getattr(p_clone, _CONTAINER_REF_ATTR, None) is not None:
                    raise RuntimeError(
                        "A Provider should always return a fresh "
                        "instance when copy() is called.")

                setattr(p_clone, _CONTAINER_REF_ATTR, ref(clone))
                clone.__providers.append(p_clone)
                clone.__singletons[type(p)] = p_clone

            return clone

    def debug(self, dependency: Hashable) -> DependencyDebug:
        with self.locked():
            for p in self.__providers:
                debug = p.maybe_debug(dependency)
                if debug is not None:
                    return debug
            raise DependencyNotFoundError(dependency)

    def provide(self, dependency: Hashable) -> DependencyValue:
        try:
            return DependencyValue(self.__singletons[dependency],
                                   scope=Scope.singleton())
        except KeyError:
            pass
        return self._safe_provide(dependency)

    def get(self, dependency: Hashable, default: object = _DEFAULT_SENTINEL) -> object:
        try:
            return self.__singletons[dependency]
        except KeyError:
            pass
        try:
            return self._safe_provide(dependency).unwrapped
        except DependencyNotFoundError:
            if default is not _DEFAULT_SENTINEL:
                return default
            raise

    def _safe_provide(self, dependency: Hashable) -> DependencyValue:
        with self._instantiation_lock:
            try:
                try:
                    return DependencyValue(self.__singletons[dependency],
                                           scope=Scope.singleton())
                except KeyError:
                    pass

                for scope, dependencies in self.__scopes.items():
                    try:
                        return DependencyValue(dependencies[dependency],
                                               scope=scope)
                    except KeyError:
                        pass

                with self._dependency_stack.instantiating(dependency):
                    for provider in self.__providers:
                        value = provider.maybe_provide(dependency, self)
                        if value is not None:
                            if value.is_singleton():
                                self.__singletons[dependency] = value.unwrapped
                            elif value.scope is not None:
                                self.__scopes[value.scope][dependency] = value.unwrapped

                            return value

            except DependencyCycleError:
                raise

            except DependencyInstantiationError as e:
                if self._dependency_stack.depth == 0:
                    raise DependencyInstantiationError(dependency) from e
                else:
                    raise

            except Exception as e:
                raise DependencyInstantiationError(
                    dependency,
                    self._dependency_stack.to_list()) from e

            raise DependencyNotFoundError(dependency)


class OverridableRawContainer(RawContainer):
    def __init__(self) -> None:
        from collections import defaultdict
        super().__init__()
        self.__override_lock = threading.RLock()
        # Used to differentiate singletons from the overrides and the "normal" ones.
        self.__singletons_override: Dict[Hashable, object] = dict()
        self.__scopes_override: Dict[Scope, Dict[Hashable, object]] = defaultdict(dict)
        self.__factory_overrides: Dict[
            Hashable, Tuple[Callable[[], object], Optional[Scope]]] = {}
        self.__provider_overrides: Deque[
            Callable[[Hashable], Optional[DependencyValue]]] = deque()

    def clone(self,
              *,
              keep_singletons: bool = False,
              keep_scopes: bool = False) -> OverridableRawContainer:
        with self.locked():
            clone = super().clone(keep_singletons=keep_singletons,
                                  keep_scopes=keep_scopes)
            if keep_singletons:
                clone.__singletons_override = self.__singletons_override
            clone.__scopes_override = {
                scope: dependencies.copy() if keep_scopes else dict()
                for scope, dependencies in self.__scopes_override.items()
            }
            clone.__factory_overrides = self.__factory_overrides
            clone.__provider_overrides = self.__provider_overrides

            return clone

    def override_singletons(self, singletons: Dict[Hashable, object]) -> None:
        with self.__override_lock:
            self.__singletons_override.update(singletons)

    def override_factory(self,
                         dependency: Hashable,
                         *,
                         factory: Callable[[], object],
                         scope: Optional[Scope]) -> None:
        with self.__override_lock:
            try:
                del self.__singletons_override[dependency]
            except KeyError:
                pass
            for scope_dependencies in self.__scopes_override.values():
                try:
                    del scope_dependencies[dependency]
                except KeyError:
                    pass
            self.__factory_overrides[dependency] = (factory, scope)

    def override_provider(self,
                          provider: Callable[[Hashable], Optional[DependencyValue]]
                          ) -> None:
        with self.__override_lock:
            self.__provider_overrides.appendleft(provider)  # latest provider wins

    def reset_scope(self, scope: Scope) -> None:
        super().reset_scope(scope)
        self.__scopes_override[scope] = dict()

    def debug(self, dependency: Hashable) -> DependencyDebug:
        from .._internal.utils.debug import debug_repr
        from .utils import DependencyDebug

        with self.__override_lock:
            try:
                value = self.__singletons_override[dependency]
            except KeyError:
                pass
            else:
                return DependencyDebug(f"Override/Singleton: {debug_repr(dependency)} "
                                       f"-> {value!r}",
                                       scope=Scope.singleton())
            try:
                (factory, scope) = self.__factory_overrides[dependency]
            except KeyError:
                pass
            else:
                return DependencyDebug(f"Override/Factory: {debug_repr(dependency)} "
                                       f"-> {debug_repr(factory)}",
                                       scope=scope)

        return super().debug(dependency)

    def provide(self, dependency: Hashable) -> DependencyValue:
        return self._safe_provide(dependency)

    def get(self, dependency: Hashable, default: object = _DEFAULT_SENTINEL) -> object:
        try:
            return self._safe_provide(dependency).unwrapped
        except DependencyNotFoundError:
            if default is not _DEFAULT_SENTINEL:
                return default
            raise

    def _safe_provide(self, dependency: Hashable) -> DependencyValue:
        with self._instantiation_lock, self.__override_lock:
            with self._dependency_stack.instantiating(dependency):
                try:
                    return DependencyValue(self.__singletons_override[dependency],
                                           scope=Scope.singleton())
                except KeyError:
                    pass

                scope: Optional[Scope]
                for scope, dependencies in self.__scopes_override.items():
                    try:
                        return DependencyValue(dependencies[dependency], scope=scope)
                    except KeyError:
                        pass

                try:
                    for provider in self.__provider_overrides:
                        value = provider(dependency)
                        if value is not None:
                            if value.scope is Scope.singleton():
                                self.__singletons_override[dependency] = value.unwrapped
                            elif value.scope is not None:
                                self.__scopes_override[value.scope][dependency] = \
                                    value.unwrapped
                            return value

                    try:
                        (factory, scope) = self.__factory_overrides[dependency]
                    except KeyError:
                        pass
                    else:
                        obj = factory()
                        if scope is Scope.singleton():
                            self.__singletons_override[dependency] = obj
                        elif scope is not None:
                            self.__scopes_override[scope][dependency] = obj
                        return DependencyValue(obj, scope=scope)

                except DependencyCycleError:
                    raise

                except Exception as e:
                    raise DependencyInstantiationError(dependency) from e

            return super()._safe_provide(dependency)



import pytest

from antidote import is_compiled


def pytest_runtest_setup(item):
    if any(mark.name == "compiled_only" for mark in item.iter_markers()):
        if not is_compiled():
            pytest.skip("Compiled only test.")



from __future__ import annotations

import functools
import inspect
import warnings
from typing import Callable, cast, Dict, Hashable, Tuple, Type, TypeVar

from typing_extensions import get_type_hints, ParamSpec

from ._internal import API
from ._internal.utils import AbstractMeta, FinalImmutable
from ._providers import FactoryProvider
from ._providers.factory import FactoryDependency
from ._providers.service import Parameterized
from ._utils import validate_method_parameters
from .core import inject
from .service import service

_ABSTRACT_FLAG = '__antidote_abstract'
P = ParamSpec('P')
T = TypeVar('T')


@API.private
class FactoryMeta(AbstractMeta):
    __factory_dependency: FactoryDependency

    def __new__(mcs: Type[FactoryMeta],
                name: str,
                bases: Tuple[type, ...],
                namespace: Dict[str, object],
                **kwargs: object
                ) -> FactoryMeta:
        abstract = kwargs.get('abstract')

        if '__call__' not in namespace and not abstract:
            raise TypeError(f"The class {name} must implement __call__()")

        cls = cast(
            FactoryMeta,
            super().__new__(mcs, name, bases, namespace, **kwargs)  # type: ignore
        )
        if not abstract:
            cls.__factory_dependency = _configure_factory(cls)

        return cls

    @API.deprecated
    def __rmatmul__(cls, left_operand: Hashable) -> object:
        warnings.warn("Prefer the Get(dependency, source=factory) notation.", DeprecationWarning)
        if left_operand is not cls.__factory_dependency.output:
            output = cls.__factory_dependency.output
            raise ValueError(f"Unsupported output {left_operand}, expected {output}")
        return cls.__factory_dependency

    @API.deprecated
    def parameterized(cls, **kwargs: object) -> PreBuild:
        """
        .. deprecated:: 1.1
            :code:`parameterized()` is a complex behavior with poor type-safety. Use-cases that
            really benefit from this behavior are few and would be better implemeneted explicitly
            in your own code.

        Creates a new dependency based on the factory with the given arguments. The new
        dependency will have the same scope as the original one.

        The recommended usage is to provide a classmethod exposing only parameters that
        may be changed:

        .. doctest:: service_meta

            >>> from antidote import Factory, world
            >>> class Database:
            ...     def __init__(self, host: str):
            ...         self.host = host
            >>> class DatabaseFactory(Factory):
            ...     __antidote__ = Factory.Conf(parameters=['host'])
            ...
            ...     def __call__(self, host: str) -> Database:
            ...         return Database(host)
            ...
            ...     @classmethod
            ...     def with_host(cls, host: str) -> object:
            ...         return cls.parameterized(host=host)
            >>> db = world.get(Database @ DatabaseFactory.with_host(host='remote'))
            >>> # or with Mypy type hint
            ... db = world.get[Database] @ DatabaseFactory.with_host(host='remote')
            >>> db.host
            'remote'
            >>> # As DatabaseFactory is defined to return a singleton,
            ... # the same is applied:
            ... world.get(Database @ DatabaseFactory.with_host(host='remote')) is db
            True

        Args:
            **kwargs: Arguments passed on to :code:`__call__()`.

        Returns:
            Dependency to be retrieved from Antidote.
        """
        warnings.warn("Deprecated, parameterized() is too complex and not type-safe",
                      DeprecationWarning)

        from .factory import Factory
        assert cls.__factory_dependency is not None

        # Guaranteed through _configure_factory()
        conf = cast(Factory.Conf, getattr(cls, '__antidote__'))
        if conf.parameters is None:
            raise RuntimeError(f"Factory {cls} does not accept any parameters. You must "
                               f"specify them explicitly in the configuration with: "
                               f"Factory.Conf(parameters=...))")

        if set(kwargs.keys()) != set(conf.parameters or []):
            raise ValueError(f"Given parameters do not match expected ones. "
                             f"Got: ({','.join(map(repr, kwargs.keys()))}) "
                             f"Expected: ({','.join(map(repr, conf.parameters))})")

        return PreBuild(cls.__factory_dependency, kwargs)


@API.private
@inject
def _configure_factory(cls: FactoryMeta,
                       factory_provider: FactoryProvider = inject.me()
                       ) -> FactoryDependency:
    from .factory import Factory

    conf = getattr(cls, '__antidote__', None)
    if not isinstance(conf, Factory.Conf):
        raise TypeError(f"Factory configuration (__antidote__) is expected to be "
                        f"a {Factory.Conf}, not a {type(conf)}")

    output = get_type_hints(cls.__call__).get('return')
    if output is None:
        raise ValueError("The return type hint is necessary on __call__."
                         "It is used a the dependency.")
    if not inspect.isclass(output):
        raise TypeError(f"The return type hint is expected to be a class, "
                        f"not {type(output)}.")

    cls = service(cls, singleton=True, wiring=conf.wiring)
    validate_method_parameters(cls.__call__, conf.parameters)

    factory_dependency = factory_provider.register(
        output=output,
        scope=conf.scope,
        factory_dependency=cls
    )

    return factory_dependency


@API.private
class FactoryWrapper:
    def __init__(self,
                 *,
                 wrapped: Callable[..., object],
                 output: object) -> None:
        self.__wrapped__ = wrapped
        self.__output = output
        functools.wraps(wrapped, updated=())(self)

    def __call__(self, *args: object, **kwargs: object) -> object:
        return self.__wrapped__(*args, **kwargs)

    def __rmatmul__(self, klass: type) -> object:
        warnings.warn("Prefer the Get(dependency, source=factory) notation.",
                      DeprecationWarning)
        if klass is not self.__output:
            raise ValueError(f"Unsupported output {klass}")
        return FactoryDependency(factory=self, output=self.__output)

    def __getattr__(self, item: str) -> object:
        return getattr(self.__wrapped__, item)

    def __repr__(self) -> str:
        return f"FactoryWrapper({self.__wrapped__!r})"


@API.private
class PreBuild(FinalImmutable):
    __slots__ = ('__factory_dependency', '__kwargs')
    __factory_dependency: FactoryDependency
    __kwargs: Dict[str, object]

    def __rmatmul__(self, left_operand: Hashable) -> object:
        if left_operand is not self.__factory_dependency.output:
            raise ValueError(f"Unsupported output {left_operand}")
        return Parameterized(self.__factory_dependency, self.__kwargs)



import functools
import inspect
import warnings
from typing import Callable, Generic, Hashable, Type, TypeVar

from typing_extensions import ParamSpec

from ._internal import API
from ._providers.indirect import ImplementationDependency

P = ParamSpec('P')
T = TypeVar('T')


@API.private
class ImplementationWrapper(Generic[P, T]):
    def __init__(self,
                 wrapped: Callable[P, T],
                 implementation_dependency: ImplementationDependency) -> None:
        self.__wrapped__ = wrapped
        self.__implementation_dependency = implementation_dependency
        functools.wraps(wrapped, updated=())(self)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return self.__wrapped__(*args, **kwargs)

    def __antidote_dependency__(self, target: Type[T]) -> object:
        if target is not self.__implementation_dependency.interface:
            interface = self.__implementation_dependency.interface
            raise ValueError(f"Unsupported interface {target}, "
                             f"expected {interface}")
        return self.__implementation_dependency

    def __rmatmul__(self, klass: type) -> object:
        warnings.warn("Prefer the Get(dependency, source=implementation) notation.",
                      DeprecationWarning)
        if klass is not self.__implementation_dependency.interface:
            interface = self.__implementation_dependency.interface
            raise ValueError(f"Unsupported interface {klass}, "
                             f"expected {interface}")
        return self.__implementation_dependency

    def __getattr__(self, item: str) -> object:
        return getattr(self.__wrapped__, item)


@API.private
def validate_provided_class(dependency: Hashable, *, expected: type) -> None:
    from ._providers.factory import FactoryDependency
    from ._providers.service import Parameterized
    from ._providers.indirect import ImplementationDependency

    cls: object = dependency
    if isinstance(cls, Parameterized):
        cls = cls.wrapped
    if isinstance(cls, FactoryDependency):
        cls = cls.output
    if isinstance(cls, ImplementationDependency):
        cls = cls.interface

    if not (isinstance(cls, type) and inspect.isclass(cls)):
        raise TypeError(f"{dependency} does not provide any class")

    if not issubclass(cls, expected):
        raise TypeError(f"Expected a subclass of {expected}, not {cls}")



from contextlib import contextmanager
from typing import Hashable, Iterator, List

from ..core.exceptions import DependencyCycleError


class DependencyStack:
    """
    Stores the stack of dependency instantiation to detect and prevent cycles
    by raising DependencyCycleError.

    Used in the Container.

    This class is not thread-safe by itself.
    """

    def __init__(self) -> None:
        self._stack: List[Hashable] = list()

    @property
    def depth(self) -> int:
        return len(self._stack)

    def to_list(self) -> List[Hashable]:
        return self._stack.copy()

    @contextmanager
    def instantiating(self, dependency: Hashable) -> Iterator[None]:
        """
        Context Manager which has to be used when instantiating the
        dependency to keep track of the dependency path.

        When a cycle is detected, a DependencyCycleError is raised.
        """
        if dependency in self._stack:
            raise DependencyCycleError(self._stack + [dependency])

        self._stack.append(dependency)
        try:
            yield
        finally:
            self._stack.pop()



from __future__ import annotations

import weakref
from typing import Callable, cast, Dict, Optional, Tuple, TYPE_CHECKING, TypeVar, Union

from typing_extensions import final

from ._internal import API
from ._internal.utils import debug_repr, FinalImmutable, short_id
from ._providers import Lazy
from .core import Container, Dependency, DependencyDebug, DependencyValue, Scope
from .service import Service

if TYPE_CHECKING:
    from .lazy import LazyMethodCall

T = TypeVar('T')


@API.private
@final
class LazyCallWithArgsKwargs(FinalImmutable, Lazy, Dependency[T]):
    """
    :meta private:
    """
    __slots__ = ('func', '_scope', '_args', '_kwargs')
    func: Callable[..., T]
    _scope: Optional[Scope]
    _args: Tuple[object, ...]
    _kwargs: Dict[str, object]

    def __antidote_debug_repr__(self) -> str:
        s = f"Lazy: {debug_repr(self.func)}(*{self._args}, **{self._kwargs})"
        if self._scope is not None:
            s += f"  #{short_id(self)}"
        return s

    def __antidote_debug_info__(self) -> DependencyDebug:
        return DependencyDebug(self.__antidote_debug_repr__(),
                               scope=self._scope,
                               wired=[self.func])

    def __antidote_provide__(self, container: Container) -> DependencyValue:
        return DependencyValue(self.func(*self._args, **self._kwargs),
                               scope=self._scope)


@API.private
@final
class LazyMethodCallWithArgsKwargs(FinalImmutable):
    """
    :meta private:
    """
    __slots__ = ('_method_name', '_scope', '_args', '_kwargs', '__cache_attr')
    _method_name: str
    _scope: Optional[Scope]
    _args: Tuple[object, ...]
    _kwargs: Dict[str, object]
    __cache_attr: str

    def __init__(self,
                 method_name: str,
                 scope: Optional[Scope],
                 args: Tuple[object, ...],
                 kwargs: Dict[str, object]
                 ) -> None:
        super().__init__(method_name, scope, args, kwargs,
                         f"__antidote_dependency_{hex(id(self))}")

    def __get__(self, instance: object, owner: type) -> object:
        if not issubclass(owner, Service):
            raise RuntimeError("LazyMethod can only be used on a Service subclass.")

        if instance is None:
            try:
                return getattr(owner, self.__cache_attr)
            except AttributeError:
                dependency = LazyMethodCallDependency(self, weakref.ref(owner))
                setattr(owner, self.__cache_attr, dependency)
                return dependency
        return getattr(instance, self._method_name)(*self._args, **self._kwargs)

    def __str__(self) -> str:
        s = f"Lazy Method: {self._method_name}(*{self._args}, **{self._kwargs})"
        if self._scope is not None:
            s += f"  #{short_id(self)}"
        return s


@API.private
@final
class LazyMethodCallDependency(FinalImmutable, Lazy):
    """
    :meta private:
    """
    __slots__ = ('__descriptor', '__owner_ref')
    __descriptor: Union[LazyMethodCall, LazyMethodCallWithArgsKwargs]
    __owner_ref: weakref.ReferenceType[type]

    def __antidote_debug_info__(self) -> DependencyDebug:
        owner = self.__owner_ref()
        assert owner is not None
        descriptor = cast('LazyMethodCall', self.__descriptor)
        return DependencyDebug(str(descriptor),
                               scope=descriptor._scope,
                               wired=[getattr(owner, descriptor._method_name)],
                               dependencies=[owner])

    def __antidote_provide__(self, container: Container) -> DependencyValue:
        owner = self.__owner_ref()
        assert owner is not None
        descriptor = cast('LazyMethodCall', self.__descriptor)
        return DependencyValue(descriptor.__get__(container.get(owner), owner),
                               scope=descriptor._scope)



from __future__ import annotations

from .. import API


@API.private
class SlotsRepr:
    __slots__ = ()

    def __repr__(self) -> str:
        slots_attrs: list[str] = []
        for cls in type(self).__mro__:
            for name in getattr(cls, '__slots__', []):
                attr = f"_{cls.__name__}{name}" if name.startswith('__') else name
                slots_attrs.append(f'{name}={getattr(self, attr)!r}')
        return f"{type(self).__name__}({', '.join(slots_attrs)})"



"""
Utilities used by world, mostly for syntactic sugar.
"""
from __future__ import annotations

import warnings
from typing import (Any, Callable, cast, Generic, Hashable, Optional, overload, Type,
                    TYPE_CHECKING,
                    TypeVar,
                    Union)

from typing_extensions import final, Protocol

from . import API
from .state import current_container
from .utils import Default, enforce_type_if_possible
from .utils.immutable import FinalImmutable, Immutable
from .utils.meta import Singleton
from ..core._annotations import extract_annotated_dependency
from ..core.annotations import Get
from ..core.container import RawContainer
from ..core.exceptions import DependencyNotFoundError
from ..core.typing import CallableClass, Dependency, Source

if TYPE_CHECKING:
    pass

T = TypeVar('T')
R = TypeVar('R')


@API.private
class SupportsRMatmul(Protocol):
    def __rmatmul__(self, type_hint: object) -> Hashable:
        pass  # pragma: no cover


@API.private
@final
class LazyDependency(Immutable, Generic[T]):
    """
    Recommended usage is to usage :py:func:`..world.lazy`:

    .. doctest:: core_utils_dependency

        >>> from antidote import Service, world
        ... class MyService(Service):
        ...     pass
        >>> port = world.lazy[MyService]()
        >>> port.get()
        <MyService ...>

    """
    __slots__ = ('unwrapped', '_type')
    unwrapped: Hashable
    """Actual dependency to be retrieved"""
    _type: Type[T]

    def __init__(self,
                 __dependency: Hashable,
                 expected_type: Type[T]) -> None:
        """
        Args:
            __dependency: actual dependency to be retrieved.
        """
        super().__init__(__dependency, expected_type)

    def get(self) -> T:
        """
        Returns:
            dependency value retrieved from :py:mod:`~..world`.
        """
        from antidote import world
        value = world.get(cast(Any, self.unwrapped))

        if not isinstance(value, self._type):
            raise TypeError(f"Dependency is not an instance of {self._type}, "
                            f"but {type(value)}")

        return value


@API.private  # use world.get, not the class directly
@final
class WorldGet(Singleton):
    @overload
    def __call__(self,
                 __dependency: Dependency[T],
                 *,
                 default: Union[T, Default] = Default.sentinel
                 ) -> T:
        ...  # pragma: no cover

    @overload
    def __call__(self,
                 __dependency: Type[T],
                 *,
                 default: Union[T, Default] = Default.sentinel
                 ) -> T:
        ...  # pragma: no cover

    @overload
    def __call__(self,
                 __dependency: Type[T],
                 *,
                 default: Union[T, Default] = Default.sentinel,
                 source: Union[Source[T], Callable[..., T], Type[CallableClass[T]]]
                 ) -> T:
        ...  # pragma: no cover

    @API.public
    def __call__(self,
                 __dependency: Any,
                 *,
                 default: Any = Default.sentinel,
                 source: Optional[Union[
                     Source[Any],
                     Callable[..., Any],
                     Type[CallableClass[Any]]
                 ]] = None
                 ) -> Any:
        try:
            __dependency = cast(Any, extract_annotated_dependency(__dependency))
            if source is not None:
                __dependency = Get(__dependency, source=source).dependency
            return current_container().get(__dependency)
        except DependencyNotFoundError:
            if default is not Default.sentinel:
                return default
            raise

    @API.public
    def __getitem__(self, tpe: Type[T]) -> TypedWorldGet[T]:
        return TypedWorldGet(tpe)


@API.private  # use world.get, not the class directly
@final
class TypedWorldGet(Generic[T], FinalImmutable):
    __slots__ = ('__type',)
    __type: Type[T]

    @overload
    def __call__(self,
                 __dependency: Any,
                 *,
                 default: Union[T, Default] = Default.sentinel
                 ) -> T:
        ...  # pragma: no cover

    @overload
    def __call__(self,
                 *,
                 default: Union[T, Default] = Default.sentinel,
                 ) -> T:
        ...  # pragma: no cover

    @overload
    def __call__(self,
                 __dependency: Type[R],
                 *,
                 default: Union[T, Default] = Default.sentinel,
                 source: Union[Source[R], Callable[..., R], Type[CallableClass[R]]]
                 ) -> T:
        ...  # pragma: no cover

    @overload
    def __call__(self,
                 *,
                 default: Union[T, Default] = Default.sentinel,
                 source: Union[Source[T], Callable[..., T], Type[CallableClass[T]]]
                 ) -> T:
        ...  # pragma: no cover

    @API.public
    def __call__(self,
                 __dependency: Any = None,
                 *,
                 default: Union[T, Default] = Default.sentinel,
                 source: Optional[Union[
                     Source[Any],
                     Callable[..., Any],
                     Type[CallableClass[Any]]
                 ]] = None
                 ) -> T:

        if default is not Default.sentinel \
                and isinstance(self.__type, type) \
                and not isinstance(default, self.__type):
            raise TypeError(f"Default value {default} is not an instance of {self.__type}, "
                            f"but a {type(default)}")
        if __dependency is None:
            __dependency = extract_annotated_dependency(self.__type)
        else:
            __dependency = extract_annotated_dependency(__dependency)
        if source is not None:
            __dependency = Get(cast(Any, __dependency), source=source).dependency
        try:
            value = cast(T, current_container().get(__dependency))
        except DependencyNotFoundError:
            if default is not Default.sentinel:
                return default
            raise

        enforce_type_if_possible(value, self.__type)
        return value

    @API.public
    def __matmul__(self, other: SupportsRMatmul) -> T:
        warnings.warn("Prefer the Get(dependency, source=X) notation.",
                      DeprecationWarning)
        return self.__call__(self.__type @ other)


@API.private
@final
class WorldLazy(Singleton):
    def __call__(self, __dependency: Hashable) -> LazyDependency[object]:
        warnings.warn("Deprecated behavior, wrap world.get() yourself", DeprecationWarning)
        return LazyDependency(__dependency, object)

    def __getitem__(self, tpe: Type[T]) -> TypedWorldLazy[T]:
        warnings.warn("Deprecated behavior, wrap world.get() yourself", DeprecationWarning)
        return TypedWorldLazy(tpe)


@API.private
@final
class TypedWorldLazy(Generic[T], Immutable):
    __slots__ = ('__type',)
    __type: Type[T]

    def __call__(self, __dependency: Hashable = None) -> LazyDependency[T]:
        warnings.warn("Deprecated behavior, wrap world.get() yourself", DeprecationWarning)
        return LazyDependency[T](self.__type if __dependency is None else __dependency,
                                 self.__type)

    def __matmul__(self, other: SupportsRMatmul) -> LazyDependency[T]:
        return self.__call__(self.__type @ other)


@API.private
def new_container() -> RawContainer:
    """ default new container in Antidote """

    from .._providers import (LazyProvider, ServiceProvider,
                              IndirectProvider, FactoryProvider)

    container = RawContainer()
    container.add_provider(FactoryProvider)
    container.add_provider(LazyProvider)
    container.add_provider(IndirectProvider)
    container.add_provider(ServiceProvider)

    return container



from __future__ import annotations

import inspect
import warnings
from typing import Any, Callable, Hashable, Optional, overload, Type, TYPE_CHECKING, TypeVar, Union

from typing_extensions import Annotated, get_type_hints, Protocol

from .marker import Marker
from .typing import CallableClass, Source
from .._internal import API
from .._internal.utils import FinalImmutable

if TYPE_CHECKING:
    from .injection import Arg


@API.private
class SupportsRMatmul(Protocol):
    def __rmatmul__(self, type_hint: object) -> Hashable:
        pass  # pragma: no cover


T = TypeVar('T')


@API.private
class AntidoteAnnotation:
    """Base class for all Antidote annotation."""
    __slots__ = ()


# API.private
INJECT_SENTINEL = AntidoteAnnotation()

# API.public
Inject = Annotated[T, INJECT_SENTINEL]
Inject.__doc__ = """
Annotation specifying that the type hint itself is the dependency:

.. doctest:: core_annotation_provide

    >>> from antidote import world, inject, Inject, service
    >>> from typing import Annotated
    ... # from typing_extensions import Annotated # Python < 3.9
    >>> @service
    ... class Database:
    ...     pass
    >>> @inject
    ... def load_db(db: Inject[Database]):
    ...     return db
    >>> load_db()
    <Database ...>

"""

# API.deprecated
Provide = Annotated[T, INJECT_SENTINEL]
Provide.__doc__ = Inject.__doc__


@API.public
class Get(FinalImmutable, AntidoteAnnotation, Marker):
    """
    Annotation specifying explicitly which dependency to inject.

    .. doctest:: core_annotation_get

        >>> from typing import Annotated
        >>> from antidote import world, inject, Get, Constants, const
        >>> class Config(Constants):
        ...     DB_HOST = const('localhost')
        >>> @inject
        ... def f(host: Annotated[str, Get(Config.DB_HOST)]):
        ...     return host
        >>> f() == world.get(Config.DB_HOST)
        True

    """
    __slots__ = ('dependency',)
    dependency: object

    @overload
    def __init__(self, __dependency: object) -> None:
        ...  # pragma: no cover

    @overload
    def __init__(self,
                 __dependency: Type[T],
                 *,
                 source: Union[Source[T], Callable[..., T], Type[CallableClass[T]]]
                 ) -> None:
        ...  # pragma: no cover

    def __init__(self,
                 __dependency: Any,
                 *,
                 source: Optional[Union[
                     Source[Any],
                     Callable[..., Any],
                     Type[CallableClass[Any]]
                 ]] = None
                 ) -> None:
        from .._providers.factory import FactoryDependency

        if isinstance(__dependency, Get):
            __dependency = __dependency.dependency

        if isinstance(source, Source):
            __dependency = source.__antidote_dependency__(__dependency)
        elif source is not None:
            if isinstance(source, type) and inspect.isclass(source):
                output = get_type_hints(source.__call__).get('return')
            elif callable(source):
                output = get_type_hints(source).get('return')
            else:
                raise TypeError(f"{source} is neither a factory function/class nor a source,"
                                f" but a {type(source)}")

            if not (isinstance(output, type) and inspect.isclass(output)):
                raise TypeError(f"{source} is not a valid factory, it must return a class")

            if not (isinstance(__dependency, type) and inspect.isclass(__dependency)):
                raise TypeError(f"dependency must be a class for a factory, "
                                f"not a {type(__dependency)}")

            if not issubclass(output, __dependency):
                raise TypeError(f"Expected dependency {__dependency} does not match output"
                                f" of the factory {source}")

            __dependency = FactoryDependency(
                factory=source,
                output=__dependency
            )

        super().__init__(__dependency)


@API.public
class From(FinalImmutable, AntidoteAnnotation):
    """
    Annotation specifying from where a dependency must be provided. To be used with
    :py:func:`~.antidote.factory`, :py:class:`.Factory` and :py:func:`.implementation`
    typically.

    .. doctest:: core_annotations_from

        >>> from typing import Annotated
        >>> from antidote import factory, world, inject, From
        >>> class Database:
        ...     def __init__(self, host: str):
        ...         self.host = host
        >>> @factory
        ... def build_db(host: str = 'localhost:5432') -> Database:
        ...     return Database(host=host)
        >>> @inject
        ... def f(db: Annotated[Database, From(build_db)]) -> Database:
        ...     return db
        >>> f().host
        'localhost:5432'

    """
    __slots__ = ('source',)
    source: SupportsRMatmul

    def __init__(self,
                 __source: Union[
                     SupportsRMatmul,
                     Source[Any],
                     Callable[..., Any],
                     Type[CallableClass[Any]]]
                 ) -> None:
        super().__init__(source=__source)


@API.deprecated
class FromArg(FinalImmutable, AntidoteAnnotation):
    """
    .. deprecated:: 1.1
        Specifying a callable to :py:func:`.inject` is deprecated, so is this annotation.
        If you rely on this behavior, you'll need to wrap @inject and do the annotation parsing
        yourself.

    Annotation specifying which dependency should be provided based on the argument. The
    function should accept a single argument of type :py:class:`~..injection.Arg` and
    return either a dependency or :py:obj:`None`.

    .. doctest:: core_annotations_from_arg

        >>> from typing import Annotated, TypeVar
        >>> from antidote import world, inject, FromArg, Constants, const, Arg
        >>> class Config(Constants):
        ...     PORT = const(5432)
        ...
        ...     @classmethod
        ...     def from_arg(cls, arg: Arg):
        ...         return getattr(cls, arg.name.upper())
        ...
        >>> T = TypeVar('T')
        >>> ProvideFromConfig = Annotated[T, FromArg(Config.from_arg)]
        >>> @inject
        ... def f(port: ProvideFromConfig[int]) -> int:
        ...     return port
        >>> f()
        5432
    """
    __slots__ = ('function',)
    function: Callable[[Arg], Optional[Hashable]]

    def __init__(self,
                 __function: Callable[[Arg], Optional[Hashable]]
                 ) -> None:
        warnings.warn("Deprecated, @inject won't support this behavior anymore", DeprecationWarning)
        if callable(__function):
            super().__init__(function=__function)
        else:
            raise TypeError(f"Expected a function, not {type(__function)}")



from contextlib import contextmanager

import pytest

from antidote import Scope
from antidote.utils import validated_scope

dummy_scope = Scope('dummy')


@contextmanager
def does_not_raise():
    yield


@pytest.mark.parametrize('expectation, kwargs', [
    pytest.param(pytest.raises(TypeError, match='.*scope.*'),
                 dict(scope=object(), default=None),
                 id='scope=object'),
    pytest.param(pytest.raises(TypeError, match='.*default.*'),
                 dict(scope=None, default=object()),
                 id='default=object'),
    pytest.param(pytest.raises(TypeError, match='.*singleton.*'),
                 dict(scope=Scope.sentinel(), singleton=object(), default=None),
                 id='singleton=object'),
    pytest.param(pytest.raises(TypeError, match='.*both.*'),
                 dict(scope=None, singleton=False, default=None),
                 id='singleton & scope'),
])
def test_invalid_validated_scope(expectation, kwargs):
    with expectation:
        assert validated_scope(**kwargs)


@pytest.mark.parametrize('scope, singleton, default, expected', [
    pytest.param(Scope.sentinel(), True, None, Scope.singleton(), id='singleton=True'),
    pytest.param(Scope.sentinel(), False, None, None, id='singleton=False'),
    pytest.param(None, None, None, None, id='scope=None'),
    pytest.param(Scope.singleton(), None, None, Scope.singleton(), id='scope=singleton'),
    pytest.param(dummy_scope, None, None, dummy_scope, id='scope=dummy'),
    pytest.param(Scope.sentinel(), None, None, None, id='default=None'),
    pytest.param(Scope.sentinel(), None, Scope.singleton(), Scope.singleton(),
                 id='default=singleton'),
    pytest.param(Scope.sentinel(), None, dummy_scope, dummy_scope, id='default=dummy'),
])
def test_validated_scope(scope, singleton, default, expected):
    assert expected == validated_scope(scope, singleton, default=default)



from typing import Type

import pytest
from typing_extensions import Annotated

from antidote import factory, Get, inject, service, world


@pytest.fixture(autouse=True)
def new_world():
    with world.test.new():
        yield


def test_injectable():
    @service
    class Dummy:
        pass

    @inject(dependencies=[Get(Dummy)])
    def f(x):
        return x

    @inject(dependencies=[Get(Dummy)])
    def g(x: Annotated[Dummy, Get(Dummy)]):
        return x

    assert f() is world.get(Dummy)
    assert g() is world.get(Dummy)


def test_nested_get():
    @service
    class Dummy:
        pass

    @inject(dependencies=[Get(Get(Dummy))])
    def f(x):
        return x

    @inject
    def g(x: Annotated[Dummy, Get(Get(Dummy))]):
        return x

    assert f() is world.get(Dummy)
    assert g() is world.get(Dummy)


def test_factory():
    class Dummy:
        pass

    dummy = Dummy()

    @factory
    def build_dummy() -> Dummy:
        return dummy

    @inject(dependencies=[Get(Dummy, source=build_dummy)])
    def f(x):
        return x

    @inject
    def g(x: Annotated[Dummy, Get(Dummy, source=build_dummy)]):
        return x

    assert f() is dummy
    assert g() is dummy


def test_source():
    class Dummy:
        pass

    class DummySource:
        SENTINEL = Dummy()

        def __antidote_dependency__(self, dependency: Type[Dummy]) -> object:
            return self.SENTINEL

    dummy_source = DummySource()
    world.test.singleton(DummySource.SENTINEL, DummySource.SENTINEL)

    @inject(dependencies=[Get(Dummy, source=dummy_source)])
    def f(x):
        return x

    @inject
    def g(x: Annotated[Dummy, Get(Dummy, source=dummy_source)]):
        return x

    assert f() is DummySource.SENTINEL
    assert g() is DummySource.SENTINEL


def test_invalid_source():
    class Dummy:
        pass

    class DummySource:
        pass

    with pytest.raises(TypeError, match=".*source.*"):
        Get(Dummy, source=DummySource())


def test_invalid_factory():
    class Dummy:
        pass

    with pytest.raises(TypeError, match=".*factory.*"):
        Get(Dummy, source=object())

    with pytest.raises(TypeError, match=".*factory.*"):
        Get(Dummy, source=Dummy)


def test_invalid_factory_return_type_hint():
    class Dummy:
        pass

    def f():
        pass

    with pytest.raises(TypeError, match=".*factory.*"):
        Get(Dummy, source=f)

    def g() -> object:
        pass

    with pytest.raises(TypeError, match=".*factory.*"):
        Get(Dummy, source=g)

    class F:
        def __call__(self):
            pass

    with pytest.raises(TypeError, match=".*valid.*factory.*return.*class"):
        Get(Dummy, source=F)

    class G:
        def __call__(self) -> object:
            pass

    with pytest.raises(TypeError, match=".*valid.*factory.*return.*class"):
        Get(Dummy, source=G)

    def dummy_factory() -> Dummy:
        return Dummy()

    with pytest.raises(TypeError, match=".*dependency.*class"):
        Get(object(), source=dummy_factory)

    with pytest.raises(TypeError, match=".*dependency.*does not match.*factory.*"):
        Get(F, source=dummy_factory)



from typing import (Any, Callable, Dict, Hashable, Optional, overload, TypeVar, Union)

from typing_extensions import get_type_hints

from ..._internal import API, state
from ...core.container import (DependencyValue, Scope)
from ...utils import validated_scope

__sentinel = object()


@overload
def singleton(dependency: Hashable, value: object) -> None:
    ...  # pragma: no cover


@overload
def singleton(dependency: Dict[Hashable, object]) -> None:
    ...  # pragma: no cover


@API.public
def singleton(dependency: Union[Dict[Hashable, object], Hashable],
              value: object = __sentinel) -> None:
    """
    Override one or multiple dependencies with one/multiple singletons.

    .. doctest:: world_test_override_singleton

        >>> from antidote import world, Service
        >>> class MyService(Service):
        ...     pass
        >>> world.get(MyService)
        <MyService ...>
        >>> with world.test.clone():
        ...     # dependencies can also be overridden as many times as necessary
        ...     world.test.override.singleton(MyService, "fake service")
        ...     # Within the cloned world, we changed the dependency value of MyService
        ...     world.get(MyService)
        'fake service'

    Args:
        dependency: Singleton to declare, must be hashable. If a dict is provided, it'll
            be treated as a dictionary of singletons to add.
        value: Associated value for the dependency.

    """
    if value is __sentinel:
        if isinstance(dependency, dict):
            state.current_overridable_container().override_singletons(dependency)
        else:
            raise TypeError("If only a single argument is provided, "
                            "it must be a dictionary of singletons.")
    else:
        if isinstance(dependency, dict):
            raise TypeError("A dictionary cannot be used as a key.")
        state.current_overridable_container().override_singletons({dependency: value})


F = TypeVar('F', bound=Callable[..., Any])


@API.public
def factory(dependency: Hashable = None,
            *,
            singleton: Optional[bool] = None,
            scope: Optional[Scope] = Scope.sentinel()
            ) -> Callable[[F], F]:
    """
    Override a dependency with the result of a factory. To be used as a function
    decorator. The result of the underlying function, the factory, will be used as the
    associated value of the dependency.

    .. doctest:: world_test_override_factory

        >>> from antidote import world, Service
        >>> class MyService(Service):
        ...     pass
        >>> world.get(MyService)
        <MyService ...>
        >>> with world.test.clone():
        ...     @world.test.override.factory(MyService)
        ...     def test():
        ...         return "fake service"
        ...     # Within the cloned world, we changed the dependency value of MyService
        ...     world.get(MyService)
        'fake service'
        >>> with world.test.clone():
        ...     # If not specified explicitly, the return type hint will be used
        ...     # as the dependency.
        ...     @world.test.override.factory()
        ...     def test() -> MyService:
        ...         return "fake service"
        ...     world.get(MyService)
        'fake service'

    Args:
        dependency: Dependency to override.
        singleton: Whether the returned dependency  is a singleton or not. If yes,
            the factory will be called at most once and the result re-used. Mutually
            exclusive with :code:`scope`. Defaults to :py:obj:`True`.
        scope: Scope of the returned dependency. Mutually exclusive with
            :code:`singleton`. The scope defines if and how long the returned dependency
            will be cached. See :py:class:`~.core.container.Scope`. Defaults to
            :py:meth:`~.core.container.Scope.singleton`.

    .. note::

        As all override work as a single layer on top of the usual dependencies, if the
        dependency is already overridden as a singleton, the factory won't be called.
        Overrides done with a provider also have a higher priority. See
        :py:mod:`.world.test.override`.

    Returns:
        The decorated function, unchanged.
    """
    scope = validated_scope(scope, singleton, default=Scope.singleton())

    def decorate(f: F) -> F:
        if not callable(f):
            raise TypeError(f"factory must be a callable, not a {type(f)}")
        if dependency is None:
            output = get_type_hints(f).get('return')
            if output is None:
                raise ValueError("Either the dependency argument or the return type hint "
                                 "of the factory must be specified")
        else:
            output = dependency
        state.current_overridable_container() \
            .override_factory(output, factory=f, scope=scope)
        return f

    return decorate


P = TypeVar('P', bound=Callable[[Any], Optional[DependencyValue]])


@API.public
def provider() -> Callable[[P], P]:
    """
    Function decorator used to declare a simplified provider to override some
    dependencies. The function will be called for each dependency and should
    return :py:class:`.core.DependencyValue` if it can be provided or :py:obj:`None`
    if not.

    .. doctest:: world_test_override_provider

        >>> from antidote import world, Service
        >>> from antidote.core import DependencyValue, Scope
        >>> class MyService(Service):
        ...     pass
        >>> world.get(MyService)
        <MyService ...>
        >>> with world.test.clone():
        ...     @world.test.override.provider()
        ...     def test(dependency):
        ...         if dependency is MyService:
        ...             return DependencyValue('fake service', scope=Scope.singleton())
        ...     # Within the cloned world, we changed the dependency value of MyService
        ...     world.get(MyService)
        'fake service'

    .. note::

        As all override work as a single layer on top of the usual dependencies, singleton
        overrides won't pass through the provider. See :py:mod:`.world.test.override`.

    .. warning::

        Beware of :py:func:`~.world.test.override.provider`, it can conflict with
        :py:func:`~.world.test.override.factory` and
        :py:func:`~.world.test.override.singleton`.
        Dependencies declared with :py:func:`~.world.test.override.singleton` will hide
        :py:func:`~.world.test.override.provider`. And
        :py:func:`~.world.test.override.provider` will hide
        :py:func:`~.world.test.override.factory`.

        Moreover it won't appear in :py:func:`.world.debug`.

    Returns:
        Function decorator.
    """

    def decorate(p: P) -> P:
        if not callable(p):
            raise TypeError(f"provider must be a callable, not {type(p)}")
        state.current_overridable_container().override_provider(p)
        return p

    return decorate



from __future__ import annotations

import pytest

from antidote import const, Constants, inject, Wiring, world
from antidote._providers import LazyProvider, ServiceProvider
from antidote.core.exceptions import DependencyInstantiationError
from antidote.exceptions import DependencyNotFoundError


class A:
    pass


@pytest.fixture(autouse=True)
def test_world():
    with world.test.empty():
        world.provider(LazyProvider)
        world.provider(ServiceProvider)
        yield


def test_simple():
    class Config(Constants):
        A = const('a')
        B = const('b')

        def provide_const(self, name, arg):
            return arg * 2

    assert world.get(Config.A) == 'aa'
    assert world.get(Config.B) == 'bb'

    conf = Config()
    assert conf.A == 'aa'
    assert conf.B == 'bb'


def test_marker():
    a = object()

    class Config(Constants):
        A = const(a)

    @inject
    def f(x: object = Config.A) -> object:
        return x

    assert f() is world.get(Config.A)


def test_default():
    a = object()
    b = object()

    class Config(Constants):
        # exists
        A = const('a')
        # default
        B = const('b', default=b)
        # nothing
        C = const('c')
        # different error
        D = const('d', default='x')

        def provide_const(self, name, arg):
            if arg == 'd':
                raise Exception()
            return dict(a=a)[arg]

    assert world.get(Config.A) is a
    assert world.get(Config.B) is b
    with pytest.raises(DependencyInstantiationError):
        world.get(Config.C)
    with pytest.raises(DependencyInstantiationError):
        world.get(Config.D)

    class Config(Constants):
        with pytest.raises(TypeError):
            X = const[A](default='x')


def test_auto_cast():
    class AutoCast(Constants):
        __antidote__ = Constants.Conf(auto_cast=True)
        A = const[int]('109')
        B = const[float]('3.14')
        C = const[str](199)

    assert AutoCast().A == 109
    assert AutoCast().B == 3.14
    assert AutoCast().C == '199'

    assert world.get(AutoCast.A) == 109
    assert world.get(AutoCast.B) == 3.14
    assert world.get(AutoCast.C) == '199'

    #######

    class LimitedAutoCast(Constants):
        __antidote__ = Constants.Conf(auto_cast=[int, float])
        A = const[int]('109')
        B = const[float]('3.14')
        C = const[str](199)

    assert LimitedAutoCast().A == 109
    assert LimitedAutoCast().B == 3.14

    with pytest.raises(TypeError, match=".*str.*"):
        _ = LimitedAutoCast().C

    assert world.get(LimitedAutoCast.A) == 109
    assert world.get(LimitedAutoCast.B) == 3.14

    with pytest.raises(DependencyInstantiationError, match=".*C.*"):
        world.get(LimitedAutoCast.C)

    #######

    class NoAutoCast(Constants):
        __antidote__ = Constants.Conf(auto_cast=False)
        A = const[int]('109')
        B = const[float]('3.14')
        C = const[str](199)

    with pytest.raises(TypeError, match=".*int.*"):
        _ = NoAutoCast().A

    with pytest.raises(TypeError, match=".*float.*"):
        _ = NoAutoCast().B

    with pytest.raises(TypeError, match=".*str.*"):
        _ = NoAutoCast().C

    #######

    class MetaDummy:
        def __new__(cls, *args, **kwargs):
            raise RuntimeError()

    class ImpossibleCast(Constants):
        __antidote__ = Constants.Conf(auto_cast=[MetaDummy])
        A = const[MetaDummy]('x')

    with pytest.raises(RuntimeError):
        _ = ImpossibleCast().A

    with pytest.raises(DependencyInstantiationError):
        world.get(ImpossibleCast.A)


def test_type_safety():
    class MetaDummy:
        def __new__(cls, *args, **kwargs):
            return object()

    class Config(Constants):
        __antidote__ = Constants.Conf(auto_cast=[MetaDummy])
        INVALID = const[int]('109')
        INVALID_CAST = const[MetaDummy]('x')

    with pytest.raises(TypeError, match=".*int.*"):
        _ = Config().INVALID

    with pytest.raises(TypeError, match=".*MetaDummy.*"):
        _ = Config().INVALID_CAST

    with pytest.raises(DependencyInstantiationError):
        world.get(Config.INVALID)

    with pytest.raises(DependencyInstantiationError):
        world.get(Config.INVALID_CAST)


def test_name():
    class Config(Constants):
        A = const()
        B = const()

        def provide_const(self, name, arg):
            return name

    assert world.get(Config.A) == 'A'
    assert world.get(Config.B) == 'B'
    conf = Config()
    assert conf.A == 'A'
    assert conf.B == 'B'


def test_default_get():
    class Config(Constants):
        NOTHING = const()
        B = const("B")

    with pytest.raises(DependencyInstantiationError):
        world.get(Config.NOTHING)

    with pytest.raises(ValueError, match=".*NOTHING.*"):
        _ = Config().NOTHING

    assert world.get(Config.B) == 'B'
    assert Config().B == "B"


def test_no_const():
    class Config(Constants):
        A = 'a'

        def provide_const(self, name, arg):
            return arg * 2

    with pytest.raises(DependencyNotFoundError):
        world.get(Config.A)

    conf = Config()
    assert conf.A == 'a'


def test_no_constants_class():
    class Config:
        A = const()

    with pytest.raises(RuntimeError, match=".*Constants.*"):
        _ = Config.A

    with pytest.raises(RuntimeError, match=".*Constants.*"):
        _ = Config().A


def test_no_get_method():
    class Config(Constants):
        A = const('a')

    assert world.get(Config.A) == 'a'


def test_invalid_conf():
    with pytest.raises(TypeError, match=".*__antidote__.*"):
        class Config(Constants):
            __antidote__ = object()


def test_no_subclass_of_constants():
    class Dummy(Constants):
        pass

    with pytest.raises(TypeError, match=".*abstract.*"):
        class SubDummy(Dummy):
            pass


@pytest.mark.parametrize('kwargs, expectation', [
    (dict(wiring=object()), pytest.raises(TypeError, match=".*wiring.*")),
    (dict(auto_cast=object()), pytest.raises(TypeError, match=".*auto_cast.*")),
    (dict(auto_cast=['1']), pytest.raises(TypeError, match=".*auto_cast.*")),
])
def test_conf_error(kwargs, expectation):
    with expectation:
        Constants.Conf(**kwargs)


@pytest.mark.parametrize('kwargs', [
    dict(wiring=Wiring(methods=['method'])),
    dict(auto_cast=frozenset((str,))),
])
def test_conf_copy(kwargs):
    conf = Constants.Conf().copy(**kwargs)
    for k, v in kwargs.items():
        assert getattr(conf, k) == v


def test_conf_repr():
    conf = Constants.Conf()
    assert "auto_cast" in repr(conf)



# flake8: noqa
# Ignoring F811 for multiple definitions
from typing import Optional

from typing_extensions import Annotated


def test_readme_simple() -> None:
    from antidote import inject, service

    @service
    class Database:
        pass

    @inject
    def f(db: Database = inject.me()) -> Database:
        return db

    assert isinstance(f(), Database)  # works !

    ######################

    f(Database())

    ######################

    from antidote import Inject

    @inject
    def f2(db: Inject[Database]) -> None:
        pass

    ######################

    @inject([Database])
    def f3(db: object) -> None:
        pass

    ######################

    @inject({'db': Database})
    def f4(db: object) -> None:
        pass

    ######################

    from typing import Optional

    class Dummy:
        pass

    # When the type_hint is optional and a marker like `inject.me()` is used, None will be
    # provided if the dependency does not exists.
    @inject
    def f5(dummy: Optional[Dummy] = inject.me()) -> Optional[Dummy]:
        return dummy

    assert f5() is None

    ######################

    from antidote import world

    # Retrieve dependencies by hand, in tests typically
    world.get(Database)
    world.get[Database](Database)  # with type hint
    world.get[Database]()  # omit dependency if it's the type hint itself

    ######################

    from antidote import service, inject

    @service(singleton=False)
    class QueryBuilder:
        # methods are also injected by default
        def __init__(self, db: Database = inject.me()) -> None:
            self._db = db

    @inject
    def load_data(builder: QueryBuilder = inject.me()) -> None:
        pass

    load_data()  # yeah !

    ######################

    from antidote import inject, Constants, const

    class Config(Constants):
        DB_HOST = const[str]('localhost')

    @inject
    def ping_db(db_host: str = Config.DB_HOST) -> None:
        pass

    ping_db()  # nice !

    ######################

    from antidote import inject, Constants, const

    class Config2(Constants):
        DB_HOST = const[str]()  # used as a type annotation
        DB_PORT = const[int]()  # and also to cast the value retrieved from `provide_const`
        # defaults are supported, used on LookupError
        DB_USER = const[str](default='postgres')

        def provide_const(self, name: str, arg: Optional[object]) -> object:
            return os.environ[name]

    import os
    os.environ['DB_HOST'] = 'localhost'
    os.environ['DB_PORT'] = '5432'

    @inject
    def check_connection(db_host: str = Config2.DB_HOST,
                         db_port: int = Config2.DB_PORT) -> None:
        pass

    check_connection()  # perfect !

    ######################

    from antidote import factory, inject

    class User:
        pass

    @factory(singleton=False)  # annotated type hints can be used or you can @inject manually
    def current_user(db: Database = inject.me()) -> User:  # return type annotation is used
        return User()

    # Note that here you *know* exactly where it's coming from.
    @inject
    def is_admin(user: User = inject.me(source=current_user)) -> None:
        pass

    ######################

    from antidote import world

    world.get(User, source=current_user)

    ######################

    from antidote import factory, inject, world

    REQUEST_SCOPE = world.scopes.new(name='request')

    @factory(scope=REQUEST_SCOPE)
    def current_user2(db: Database = inject.me()) -> User:
        return User()

    # Reset all dependencies in the specified scope.
    world.scopes.reset(REQUEST_SCOPE)


def test_interface_impl() -> None:
    from antidote import implementation, service, factory, Get

    class Cache:
        pass

    @service
    class MemoryCache(Cache):
        pass

    class Redis:
        """ class from an external library """

    @factory
    def redis_cache() -> Redis:
        return Redis()

    @implementation(Cache)
    def cache_impl() -> object:
        import os

        if os.environ.get('USE_REDIS_CACHE'):
            return Get(Redis, source=redis_cache)

        # Returning the dependency that must be retrieved
        return MemoryCache

    ######################

    from antidote import world, inject

    @inject
    def heavy_compute(cache: Cache = inject.me(source=cache_impl)) -> None:
        pass

    world.get(Cache, source=cache_impl)


def test_debugging() -> None:
    from antidote import service, inject

    @service
    class Database:
        pass

    @inject
    def f(db: Database = inject.me()) -> None:
        pass

    f()
    f(Database())  # test with specific arguments in unit tests

    ######################

    from antidote import world

    # Clone current world to isolate it from the rest
    with world.test.clone():
        x = object()
        # Override the Database
        world.test.override.singleton(Database, x)
        f()  # will have `x` injected for the Database

        @world.test.override.factory(Database)
        def override_database() -> object:
            class DatabaseMock:
                pass

            return DatabaseMock()

        f()  # will have `DatabaseMock()` injected for the Database


def test_readme() -> None:
    # from a library
    class ImdbAPI:
        def __init__(self, host: str, port: int, api_key: str):
            pass

    ######################

    # movie.py
    class MovieDB:
        """ Interface """

        def get_best_movies(self) -> None:
            pass

    ######################

    # config.py
    from antidote import Constants, const

    class Config(Constants):
        # with str/int/float, the type hint is enforced. Can be removed or extend to
        # support Enums.
        IMDB_HOST = const[str]('imdb.host')
        IMDB_PORT = const[int]('imdb.port')
        IMDB_API_KEY = const[str]('imdb.api_key')

        def __init__(self) -> None:
            self._raw_conf = {
                'imdb': {
                    'host': 'dummy_host',
                    'api_key': 'dummy_api_key',
                    'port': '80'
                }
            }

        def provide_const(self, name: str, arg: Optional[str]) -> object:
            assert arg is not None
            root, key = arg.split('.')
            return self._raw_conf[root][key]

    ######################

    # current_movie.py
    # Code implementing/managing MovieDB
    from antidote import factory, inject, Service, implementation
    # from config import Config

    # Provides ImdbAPI, as defined by the return type annotation.
    @factory
    @inject([Config.IMDB_HOST, Config.IMDB_PORT, Config.IMDB_API_KEY])
    def imdb_factory(host: str, port: int, api_key: str) -> ImdbAPI:
        # Here host = Config().provide_const('IMDB_HOST', 'imdb.host')
        return ImdbAPI(host=host, port=port, api_key=api_key)

    class IMDBMovieDB(MovieDB, Service):
        __antidote__ = Service.Conf(singleton=False)  # New instance each time

        @inject({'imdb_api': ImdbAPI @ imdb_factory})
        def __init__(self, imdb_api: ImdbAPI) -> None:
            self._imdb_api = imdb_api

        def get_best_movies(self) -> None:
            pass

    @implementation(MovieDB)
    def current_movie_db() -> object:
        return IMDBMovieDB  # dependency to be provided for MovieDB

    ######################

    # current_movie.py
    # Code implementing/managing MovieDB
    from antidote import factory, Service, Get, From
    # from typing import Annotated
    # # from typing_extensions import Annotated # Python < 3.9
    # from config import Config

    @factory
    def imdb_factory2(host: Annotated[str, Get(Config.IMDB_HOST)],
                      port: Annotated[int, Get(Config.IMDB_PORT)],
                      api_key: Annotated[str, Get(Config.IMDB_API_KEY)]
                      ) -> ImdbAPI:
        return ImdbAPI(host, port, api_key)

    class IMDBMovieDB2(MovieDB, Service):
        __antidote__ = Service.Conf(singleton=False)

        def __init__(self, imdb_api: Annotated[ImdbAPI, From(imdb_factory2)]):
            self._imdb_api = imdb_api

        def get_best_movies(self) -> None:
            pass

    ######################

    # main.py
    # from movie import MovieDB
    # from current_movie import current_movie_db

    @inject([MovieDB @ current_movie_db])
    def main(movie_db: Optional[MovieDB] = None) -> None:
        assert movie_db is not None  # for Mypy, to understand that movie_db is optional
        pass

    # Or with annotated type hints
    @inject
    def main2(movie_db: Annotated[MovieDB, From(current_movie_db)]) -> None:
        pass

    main2()

    ######################

    conf = Config()
    main2(IMDBMovieDB(imdb_factory(
        # constants can be retrieved directly on an instance
        host=conf.IMDB_HOST,
        port=conf.IMDB_PORT,
        api_key=conf.IMDB_API_KEY,
    )))

    ######################

    from antidote import world

    # Clone current world to isolate it from the rest
    with world.test.clone():
        # Override the configuration
        world.test.override.singleton(Config.IMDB_HOST, 'other host')
        main2()

    ######################

    world.debug(main2)



from typing import Optional

import pytest

from antidote import factory, inject, service, world
from antidote.core.marker import Marker
from antidote.exceptions import DependencyNotFoundError


def test_marker_me():
    with world.test.new():
        @service
        class MyService:
            pass

        @inject
        def f(my_service: MyService = inject.me()):
            return my_service

        assert f() is world.get[MyService]()


def test_invalid_marker_me_missing_dependency():
    class MyService:
        pass

    @inject
    def f(my_service: MyService = inject.me()):
        return my_service

    # Should not fail before.
    with pytest.raises(DependencyNotFoundError):
        f()


def test_invalid_marker_me_wrong_type_hint():
    with pytest.raises(TypeError, match=r"(?i).*inject\.me.*"):
        @inject
        def f(my_service=inject.me()):
            return my_service

    with pytest.raises(TypeError, match=r"(?i).*inject\.me.*"):
        @inject
        def g(my_service: int = inject.me()):
            return my_service


def test_marker_me_from():
    with world.test.new():
        class MyService:
            pass

        @factory
        def create_service() -> MyService:
            return MyService()

        @inject
        def f(my_service: MyService = inject.me(source=create_service)):
            return my_service

        assert f() is world.get(MyService, source=create_service)


def test_invalid_marker_me_from_wrong_type_hint():
    with world.test.new():
        class MyService:
            pass

        @factory
        def create_service() -> MyService:
            return MyService()

        with pytest.raises(TypeError, match=r"(?i).*inject\.me.*"):
            @inject
            def f(my_service=inject.me(source=create_service)):
                return my_service

        with pytest.raises(TypeError, match=r"(?i).*inject\.me.*"):
            @inject
            def g(my_service: int = inject.me(source=create_service)):
                return my_service

        class OtherService:
            pass

        with pytest.raises(TypeError, match=r"(?i).*does not match.*"):
            @inject
            def h(my_service: OtherService = inject.me(source=create_service)):
                return my_service


def test_marker_get():
    with world.test.new():
        @service
        class MyService:
            pass

        @inject
        def f(my_service=inject.get(MyService)):
            return my_service

        assert f() is world.get[MyService]()


def test_invalid_marker_get_missing_dependency():
    class MyService:
        pass

    @inject
    def f(my_service=inject.get(MyService)):
        return my_service

    # Should not fail before.
    with pytest.raises(DependencyNotFoundError):
        f()


def test_marker_from_get():
    with world.test.new():
        class MyService:
            pass

        @factory
        def create_service() -> MyService:
            return MyService()

        @inject
        def f(my_service=inject.get(MyService, source=create_service)):
            return my_service

        assert f() is world.get(MyService, source=create_service)


def test_invalid_marker_from_get_wrong_target():
    with world.test.new():
        class MyService:
            pass

        @factory
        def create_service() -> MyService:
            return MyService()

        class OtherService:
            pass

        with pytest.raises(TypeError, match=r"(?i).*does not match.*"):
            @inject
            def f(my_service: OtherService = inject.me(source=create_service)):
                return my_service


def test_custom_marker():
    class CustomMarker(Marker):
        pass

    with pytest.raises(TypeError, match="(?i).*custom marker.*"):
        @inject
        def test(test=CustomMarker()):
            pass


def test_marker_me_optional():
    with world.test.new():
        @service
        class MyService:
            pass

        @inject
        def f(my_service: Optional[MyService] = inject.me()):
            return my_service

        assert f() is world.get[MyService]()

        with world.test.empty():
            assert f() is None

    with world.test.new():
        class MyService:
            pass

        @factory
        def create_service() -> MyService:
            return MyService()

        @inject
        def f(my_service: Optional[MyService] = inject.me(source=create_service)):
            return my_service

        assert f() is world.get(MyService, source=create_service)

        with world.test.empty():
            assert f() is None



import inspect
import warnings
from typing import Any, cast, Hashable, Type, TypeVar

from .._internal import API
from .._internal.state import current_container, init
from .._internal.world import WorldGet, WorldLazy
from ..core.container import RawProvider, Scope

init()

__sentinel = object()
T = TypeVar('T')

# API.public
get = WorldGet()
get.__doc__ = """
Used to retrieve a dependency from Antidote. A type hint can be provided and the result
will cast to match it. It follows the same philosophy as mypy and will NOT enforce it.

.. doctest:: world_get

    >>> from antidote import world, service
    >>> @service
    ... class Dummy:
    ...     pass
    >>> world.get(Dummy)
    <Dummy ...>
    >>> # But Mypy will only an object, not a `Dummy` instance. For this Mypy needs help:
    >>> world.get[Dummy](Dummy)  # Treated by Mypy as a Dummy
    <Dummy ...>
    >>> # To avoid repetition, if a type hint is provided but no argument, it
    ... # will retrieve the type hint itself.
    ... world.get[Dummy]()
    <Dummy ...>

"""


def __apply_lazy(_: object) -> WorldLazy:
    return WorldLazy()


@API.deprecated
@__apply_lazy
def lazy() -> Any:
    warnings.warn("Deprecated behavior, wrap world.get() yourself",
                  DeprecationWarning)  # pragma: no cover


lazy.__doc__ = """
.. deprecated:: 1.1
    If you need lazy behavior, wrap world.get() yourself. This feature barely brings anything.

Used to retrieves lazily a dependency. A type hint can be provided and the retrieved
instance will be cast to match it. It follows the same philosophy as mypy and will NOT
enforce it.

.. doctest:: world_lazy

    >>> from antidote import world, Service
    >>> class Dummy(Service):
    ...     pass
    >>> dep = world.lazy(Dummy)
    >>> dep.get()
    <Dummy ...>
    >>> # But Mypy will only an object, not a `Dummy` instance. For this Mypy needs help:
    >>> world.lazy[Dummy](Dummy).get()  # Treated by Mypy as a Dummy
    <Dummy ...>
    >>> # To avoid repetition, if a type hint is provided but no argument, it
    ... # will retrieve the type hint itself.
    ... world.lazy[Dummy]().get()
    <Dummy ...>

"""

P = TypeVar('P', bound=Type[RawProvider])


@API.public
def provider(p: P) -> P:
    """
    Class decorator used to register a new
    :py:class:`~.core.provider.Provider`.

    .. doctest:: world_provider

        >>> from antidote import world
        >>> from antidote.core import Provider
        >>> @world.provider
        ... class MyProvider(Provider):
        ...     '''Implementation missing'''

    Args:
        p: Provider class

    Returns:
        The same class, unmodified.

    """
    if not (inspect.isclass(p) and issubclass(p, RawProvider)):
        raise TypeError(f"Provider must be a subclass of "
                        f"RawProvider, not {p}")
    container = current_container()
    if any(p == type(existing_provider) for existing_provider in container.providers):
        raise ValueError(f"Provider {p} already exists")
    container.add_provider(p)
    return cast(P, p)


@API.public
def freeze() -> None:
    """
    Freezes Antidote. No additional dependencies or scope can be defined.

    Its primary purpose is to state explicitly in your code when all dependencies have
    been defined and to offer a bit more control on Antidote's state.

    .. doctest:: world_freeze

        >>> from antidote import world, Service
        >>> world.freeze()
        >>> class Dummy(Service):
        ...     pass
        Traceback (most recent call last):
        ...
        FrozenWorldError

    """
    current_container().freeze()


@API.public
def debug(dependency: Hashable, *, depth: int = -1) -> str:
    """
    To help you debug issues you may encounter with your injections, this will provide
    a tree representing all the dependencies that will be retrieved by Antidote when
    instantiating the specified dependency. It will also show whether those are singletons
    or if there are any cyclic dependencies.

    .. doctest:: world_debug

        >>> from antidote import world, Service, Provide
        >>> class Dummy(Service):
        ...     pass
        >>> class Root(Service):
        ...     def __init__(self, dummy: Provide[Dummy]):
        ...         pass
        >>> print(world.debug(Root))
        Root
        └── Dummy
        <BLANKLINE>
        Singletons have no scope markers.
        <∅> = no scope (new instance each time)
        <name> = custom scope
        <BLANKLINE>

    .. note::

        The output is not part of the public API. Any variation in it will not be
        considered as a breaking change. So only rely on it to actively debug issues.

    .. warning::

        With :py:func:`.implementation`, it'll execute the function to know which
        dependency to follow.

    Args:
        dependency: Root of the dependency tree.
        depth: Maximum depth of the result tree. Defaults to -1 which implies not limit.

    Returns:

    """
    from .._internal.utils.debug import tree_debug_info

    return tree_debug_info(current_container(), dependency, depth)


@API.public
def new(*, name: str) -> Scope:
    """
    Creates a new scope. See :py:class:`~.core.container.Scope` for more information on
    scopes.

    .. doctest:: world_scopes_new

        >>> from antidote import world, Service
        >>> REQUEST_SCOPE = world.scopes.new(name='request')
        >>> class Dummy(Service):
        ...     __antidote__ = Service.Conf(scope=REQUEST_SCOPE)

    Args:
        name: Friendly identifier used for debugging purposes. It must be unique.
    """
    from .._internal.state import current_container
    if not isinstance(name, str):
        raise TypeError(f"name must be a str, not {type(name)}")
    if not name:
        raise ValueError("name cannot be an empty string")
    if name in {Scope.singleton().name, Scope.sentinel().name}:
        raise ValueError(f"'{name}' is a reserved scope name.")
    container = current_container()
    if any(s.name == name for s in container.scopes):
        raise ValueError(f"A scope '{name}' already exists")
    return container.create_scope(name)


@API.public
def reset(scope: Scope) -> None:
    """
    All dependencies values of the specified scope will be discarded, so invalidating the
    scope. See :py:class:`~.core.container.Scope` for more information on scopes.

    .. doctest:: world_scopes_reset

        >>> from antidote import world
        >>> REQUEST_SCOPE = world.scopes.new(name='request')
        >>> # All cached dependencies value with scope=REQUEST_SCOPE will be discarded.
        ... world.scopes.reset(REQUEST_SCOPE)

    Args:
        scope: Scope to reset.
    """
    if not isinstance(scope, Scope):
        raise TypeError(f"scope must be a Scope, not {type(scope)}.")
    if scope in {Scope.singleton(), Scope.sentinel()}:
        raise ValueError(f"Cannot reset {scope}.")
    container = current_container()
    if scope not in container.scopes:
        raise ValueError(f"Unknown scope {scope}. Only scopes created through "
                         f"world.scopes.new() are supported.")
    return container.reset_scope(scope)



from __future__ import annotations

from typing import (Any, cast, Dict, Generic, Hashable, Optional, Tuple, Type,
                    TYPE_CHECKING, TypeVar, Union)

from typing_extensions import final

from ._internal import API
from ._internal.utils import AbstractMeta, debug_repr, Default, enforce_type_if_possible, \
    FinalImmutable, FinalMeta
from ._internal.utils.immutable import Immutable
from ._providers.lazy import Lazy
from .core import Container, DependencyDebug, DependencyValue, Scope
from .core.marker import Marker
from .core.typing import Dependency

if TYPE_CHECKING:
    from .constants import Constants

T = TypeVar('T')
Tco = TypeVar('Tco', covariant=True)


@API.private
@final
class MakeConst(metaclass=FinalMeta):
    def __call__(self,
                 __arg: Optional[object] = None,
                 *,
                 default: object = Default.sentinel) -> LazyConstDescriptor[object]:
        # Not true yet, but will be changed by ConstantsMeta / @constants
        return cast(LazyConstDescriptor[object], LazyConstToDo(__arg, None, default))

    def __getitem__(self, tpe: Type[T]) -> MakeTypedConst[T]:
        return MakeTypedConst(tpe)


@API.private
@final
class MakeTypedConst(Immutable, Generic[T]):
    __slots__ = ('__type',)
    __type: Type[T]

    def __call__(self,
                 __arg: Optional[object] = None,
                 *,
                 default: Union[T, Default] = Default.sentinel) -> LazyConstDescriptor[T]:
        if default is not Default.sentinel:
            enforce_type_if_possible(default, self.__type)
        # Not true yet, but will be changed by ConstantsMeta / @constants
        return cast(LazyConstDescriptor[T], LazyConstToDo(__arg, self.__type, default))


@API.private
@final
class LazyConstToDo(FinalImmutable):
    __slots__ = ('arg', 'type_', 'default')
    arg: Optional[object]
    type_: Optional[type]
    default: object

    def __get__(self, instance: object, owner: object) -> None:
        raise RuntimeError("const() can only be used in a subclass of Constants.")


@API.private
class ConstantsMeta(AbstractMeta):
    def __new__(mcs: Type[ConstantsMeta],
                name: str,
                bases: Tuple[type, ...],
                namespace: Dict[str, object],
                **kwargs: object
                ) -> ConstantsMeta:
        cls = cast(
            ConstantsMeta,
            super().__new__(mcs, name, bases, namespace, **kwargs)  # type: ignore
        )
        if not kwargs.get('abstract'):
            _configure_constants(cls)
        return cls


@API.private
def _configure_constants(cls: type, conf: object = None) -> None:
    from .constants import Constants
    from .service import service

    conf = conf or getattr(cls, '__antidote__', None)
    if not isinstance(conf, Constants.Conf):
        raise TypeError(f"Constants configuration (__antidote__) is expected to be a "
                        f"{Constants.Conf}, not a {type(conf)}")

    cls = service(cls, singleton=True, wiring=conf.wiring)
    for name, v in list(cls.__dict__.items()):
        if not isinstance(v, LazyConstToDo):
            continue

        descriptor: LazyConstDescriptor[Any] = LazyConstDescriptor(
            name=name,
            dependency=cls,
            method_name=Constants.provide_const.__name__,
            arg=v.arg,
            default=v.default,
            type_=v.type_ or object,
            auto_cast=v.type_ is not None and v.type_ in conf.auto_cast
        )
        setattr(cls, name, descriptor)


@API.private
@final
class LazyConstDescriptor(Generic[Tco], FinalImmutable):
    __slots__ = ('name', 'dependency', 'method_name', 'arg', 'default', 'type_',
                 'auto_cast', '_cache')
    name: str
    dependency: Hashable
    method_name: str
    arg: object
    default: Tco
    type_: type
    auto_cast: bool
    _cache: str

    def __init__(self,
                 *,
                 name: str,
                 dependency: Hashable,
                 method_name: str,
                 arg: object,
                 default: Tco,
                 type_: type,
                 auto_cast: bool
                 ):
        super().__init__(
            name=name,
            dependency=dependency,
            method_name=method_name,
            arg=arg,
            default=default,
            type_=type_,
            auto_cast=auto_cast,
            _cache=f"__antidote_dependency_{hex(id(self))}"
        )

    def __get__(self,
                instance: Optional[Constants],
                owner: Type[Constants]
                ) -> Tco:  # Not true for class instance, but helps with typing errors.
        if instance is None:
            try:
                return cast(Tco, getattr(owner, self._cache))
            except AttributeError:
                dependency = LazyConst(self)
                setattr(owner, self._cache, dependency)

                # Lying to Mypy, but it should help to detect errors when used as a Marker.
                return cast(Tco, dependency)
        try:
            value = getattr(instance, self.method_name)(name=self.name,
                                                        arg=self.arg)
        except LookupError:
            if self.default is not Default.sentinel:
                return self.default
            raise

        if self.auto_cast:
            value = self.type_(value)

        enforce_type_if_possible(value, self.type_)
        return cast(Tco, value)


@API.private
@final
class LazyConst(FinalImmutable, Lazy, Dependency[T], Marker):
    __slots__ = ('descriptor',)
    descriptor: object

    def __init__(self, descriptor: LazyConstDescriptor[T]) -> None:
        super().__init__(descriptor=descriptor)

    def __antidote_debug_info__(self) -> DependencyDebug:
        descriptor = cast(LazyConstDescriptor[T], self.descriptor)
        cls = cast(type, descriptor.dependency)
        return DependencyDebug(f"{debug_repr(cls)}.{descriptor.name}",
                               scope=Scope.singleton(),
                               # TODO: Would be great if the first argument of the method
                               #       didn't show as unknown as it's always provided.
                               wired=[getattr(cls, descriptor.method_name), cls])

    def __antidote_provide__(self, container: Container) -> DependencyValue:
        # TODO: Waiting for a fix: https://github.com/python/mypy/issues/6910
        descriptor = cast(LazyConstDescriptor[T], self.descriptor)
        return DependencyValue(
            getattr(container.get(descriptor.dependency), descriptor.name),
            scope=Scope.singleton()
        )



from typing import Hashable, Optional, Sequence

from typing_extensions import final

from .container import Scope
from .._internal import API
from .._internal.utils import FinalImmutable


@API.public
@final
class DependencyDebug(FinalImmutable):
    """
    Debug information on a dependency. Used by :py:mod:`.world.debug` to provide runtime
    information for debugging.
    """
    __slots__ = ('info', 'scope', 'wired', 'dependencies')
    info: str
    scope: Optional[Scope]
    wired: Sequence[object]
    dependencies: Sequence[Hashable]

    def __init__(self,
                 __info: str,
                 *,
                 scope: Optional[Scope] = None,
                 wired: Sequence[object] = tuple(),
                 dependencies: Sequence[Hashable] = tuple()):
        """
        Args:
            __info: Short and concise information on the dependency, just enough to
                identify clearly which one it is.
            scope: Scope of the dependency.
            wired: Every class or function that may have been wired for this dependency.
            dependencies: Any transient dependency, so dependencies of this dependency.
        """
        super().__init__(__info, scope, wired, dependencies)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, DependencyDebug) \
               and self.info == other.info \
               and self.scope == other.scope \
               and self.wired == other.wired \
               and self.dependencies == other.dependencies  # noqa: E126



from __future__ import annotations

import collections.abc as c_abc
from typing import (Any, FrozenSet, Iterable, Optional, Sequence, Union, ClassVar)

from typing_extensions import final

from ._constants import ConstantsMeta, MakeConst
from ._internal import API
from ._internal.utils import Copy, FinalImmutable
from .core.wiring import Wiring, WithWiringMixin

# API.public
const = MakeConst()


@API.public
class Constants(metaclass=ConstantsMeta, abstract=True):
    """
    Used to declare constants, typically configuration which may be retrieved from the
    environment, a file, etc...

    Constants are simply defined with :py:func:`.const`:

    .. doctest:: constants

        >>> from antidote import Constants, const, world
        >>> class Config(Constants):
        ...     PORT = const(80)
        ...     HOST = const('localhost')
        >>> world.get[int](Config.PORT)
        80
        >>> world.get[str](Config.HOST)
        'localhost'
        >>> # Or directly from an instance, for testing typically:
        ... Config().PORT
        80

    :py:class:`.Constants` really shines when you need to change *how* you retrieve the
    constants. For example, let's suppose we need to load our configuration from a file:

    .. doctest:: constants

        >>> class Config(Constants):
        ...     PORT = const[int]('port')
        ...     WRONG_TYPE = const[Constants]('port')
        ...     HOST = const[str]('host')
        ...
        ...     def __init__(self):
        ...         self._data = {'host': 'localhost', 'port': '80'}
        ...
        ...     def provide_const(self, name: str, arg: str):
        ...         return self._data[arg]
        ...
        >>> # the actual constant will be auto casted to the type hint if present and
        ... # is one of {str, float, int}. In all cases, the type of the constant value
        ... # is always checked at runtime.
        ... world.get(Config.PORT)
        80
        >>> Config().WRONG_TYPE
        Traceback (most recent call last):
          File "<stdin>", line 1, in ?
        TypeError
        >>> # The type hint is also used as... a type hint !
        ... Config().PORT  # will be treated as an int by Mypy
        80
        >>> world.get(Config.PORT)  # will also be treated as an int by Mypy
        80
        >>> world.get(Config.HOST)
        'localhost'

    In the previous example we're also using the :code:`auto_cast` feature. It provides
    a type hint for Mypy and will be used to cast the result of
    :py:meth:`.Constants.provide_const` if one of :code:`str`, :code:`int` or
    :code:`float`. It can be configured to be either deactivated or extended to support
    enums for example.

    Another useful feature is :code:`default` which simply defines the default value to
    be used if a :py:exc:`LookUpError` is raised in :py:meth:`.Constants.provide_const`.
    It must already have the correct type, :code:`auto_cast` will *not* be applied on it.

    .. doctest:: constants

        >>> class Config(Constants):
        ...     PORT = const[int]('port', default=80)
        ...     HOST = const[str]('host')
        ...
        ...     def __init__(self):
        ...         self._data = {'host': 'localhost'}
        ...
        ...     def provide_const(self, name: str, arg: str):
        ...         return self._data[arg]
        ...
        >>> world.get(Config.PORT)
        80
        >>> world.get(Config.HOST)
        'localhost'

    """
    __slots__ = ()

    @final
    class Conf(FinalImmutable, WithWiringMixin):
        """
        Immutable Constants configuration. To change parameters on a existing instance,
        use either method :py:meth:`.copy` or
        :py:meth:`~.core.wiring.WithWiringMixin.with_wiring`.
        """
        __slots__ = ('wiring', 'auto_cast')
        wiring: Optional[Wiring]
        auto_cast: FrozenSet[type]

        def __init__(self,
                     *,
                     auto_cast: Union[Iterable[type], bool] = True,
                     wiring: Optional[Wiring] = Wiring()):
            """
            Args:
                wiring: :py:class:`Wiring` used on the class. Defaults to wire only
                    :code:`__init__()`.
                auto_cast: When the type of the constant is specified with
                    :py:func:`.const`, the value of the dependency will be cast to its
                    type if it's one of :code:`str`, :code:`float` or :code:`int` by
                    default. You can disable this by specifying code:`auto_cast=False` or
                    change the types for which it's done by specifying explicitly those
                    types.
            """
            if not (wiring is None or isinstance(wiring, Wiring)):
                raise TypeError(f"wiring can be None or a Wiring, "
                                f"but not a {type(wiring)}")

            if isinstance(auto_cast, bool):
                if auto_cast:
                    auto_cast = frozenset((float, int, str))
                else:
                    auto_cast = frozenset()
            elif isinstance(auto_cast, c_abc.Iterable):
                auto_cast = frozenset(auto_cast)
                if not all(isinstance(tpe, type) for tpe in auto_cast):
                    raise TypeError("auto_cast must a boolean or a iterable of types")
            else:
                raise TypeError("auto_cast must a boolean or a iterable of types")

            super().__init__(wiring=wiring, auto_cast=auto_cast)

        def copy(self,
                 *,
                 wiring: Union[Optional[Wiring], Copy] = Copy.IDENTICAL,
                 auto_cast: Union[Union[Sequence[type], bool], Copy] = Copy.IDENTICAL
                 ) -> Constants.Conf:
            """
            Copies current configuration and overrides only specified arguments.
            Accepts the same arguments as :py:meth:`.__init__`
            """
            return Copy.immutable(self, wiring=wiring, auto_cast=auto_cast)

    __antidote__: ClassVar[Conf] = Conf()

    def provide_const(self, *, name: str, arg: Optional[Any]) -> object:
        """
        Used to retrieve the value of the constant defined with :py:func:`.const`.
        If a :py:exc:`LookUpError` is raised, the :code:`default` value defined
        with :py:func:`.const` will be used if any.

        Args:
            name: Name of the constant
            arg: Argument given to :py:func:`.const` when creating the constant.

        Returns:
            Constant value.
        """
        if arg is None:
            raise ValueError(f"No value provided for const {name}")
        return arg



import pytest

from antidote import wire, world


@pytest.fixture(autouse=True)
def new_world():
    with world.test.empty():
        yield


@pytest.mark.parametrize('obj', [object(), lambda: None])
def test_invalid_class(obj):
    with pytest.raises(TypeError):
        wire(obj)


@pytest.mark.parametrize(
    'kwargs',
    [dict(methods=object()),
     dict(auto_provide=object()),
     dict(dependencies=object())]
)
def test_invalid_type(kwargs):
    with pytest.raises(TypeError):
        @wire(**kwargs)
        class Dummy:
            pass
